"""animation loading and manipulating functions.

<p>please note that this file is alpha, and is subject to modification in
future versions of pgu!</p>
"""

print("pgu.ani", "This module is alpha, and is subject to change.")

import math
import pygame


def _ani_load(tv, name, parts, frames, shape):
    l = len(frames)
    # print name,parts,l
    n = parts.pop()
    if len(parts):
        s = l / n
        for i in range(0, n):
            _ani_load(
                tv, name + ".%d" % i, parts[:], frames[s * i : s * (i + 1)], shape
            )
        return

    for i in range(0, n):
        tv.images[name + ".%d" % i] = frames[i], shape


def ani_load(tv, name, img, size, shape, parts):
    """load an animation from an image

    <pre>ani_load(tv,name,image,size,shape,parts)</pre>

    <dl>
    <dt>tv<dd>vid to load into
    <dt>name <dd>prefix name to give the images
    <dt>image <dd>image to load anis from
    <dt>size <dd>w,h size of image
    <dt>shape <dd>shape of image (usually a subset of 0,0,w,h) used for collision detection
    <dt>parts <dd>list of parts to divide the animation into
        <br>for example parts = [4,5] would yield 4 animations 5 frames long, 20 total
        <br>for example parts = [a,b,c] would yield ... images['name.a.b.c'] ..., a*b*c total
    </dl>

    """
    parts = parts[:]
    parts.reverse()
    w, h = size
    frames = []
    for y in range(0, img.get_height(), h):
        for x in range(0, img.get_width(), w):
            frames.append(img.subsurface(x, y, w, h))
    _ani_load(tv, name, parts, frames, shape)


def image_rotate(tv, name, img, shape, angles, diff=0):
    """rotate an image and put it into tv.images

    <pre>image_rotate(tv,name,image,shape,angles,diff=0)</pre>

    <dl>
    <dt>tv <dd>vid to load into
    <dt>name <dd>prefix name to give the images
    <dt>image <dd>image to load anis from
    <dt>shape <dd>shape fimage (usually a subset of 0,0,w,h) used for collision detection
    <dt>angles <dd>a list of angles to render in degrees
    <dt>diff <dd>a number to add to the angles, to correct for source image not actually being at 0 degrees
    </dl>
    """
    w1, h1 = img.get_width(), img.get_height()
    shape = pygame.Rect(shape)
    ps = shape.topleft, shape.topright, shape.bottomleft, shape.bottomright
    for a in angles:
        img2 = pygame.transform.rotate(img, a + diff)
        w2, h2 = img2.get_width(), img2.get_height()
        minx, miny, maxx, maxy = 1024, 1024, 0, 0
        for x, y in ps:
            x, y = x - w1 / 2, y - h1 / 2
            a2 = math.radians(a + diff)
            # NOTE: the + and - are switched from the normal formula because of
            # the weird way that pygame does the angle...
            x2 = x * math.cos(a2) + y * math.sin(a2)
            y2 = y * math.cos(a2) - x * math.sin(a2)
            x2, y2 = x2 + w2 / 2, y2 + h2 / 2
            minx = min(minx, x2)
            miny = min(miny, y2)
            maxx = max(maxx, x2)
            maxy = max(maxy, y2)
        r = pygame.Rect(minx, miny, maxx - minx, maxy - miny)
        # print r
        # ((ww-w)/2,(hh-h)/2,w,h)
        tv.images["%s.%d" % (name, a)] = img2, r



import pygame
from pygame.locals import *

pygame.display.init()
pygame.display.set_mode((80, 80), 32)


def prep(name):
    fname = name + ".png"
    img = pygame.image.load(fname)
    w, h = img.get_width() / 2, img.get_height() / 2

    out = pygame.Surface((w * 3, h * 3), SWSURFACE | SRCALPHA, 32)
    out.fill((0, 0, 0, 0))
    out.blit(img.subsurface(0, 0, w, h), (0, 0))
    out.blit(img.subsurface(w, 0, w, h), (w * 2, 0))
    out.blit(img.subsurface(0, h, w, h), (0, h * 2))
    out.blit(img.subsurface(w, h, w, h), (w * 2, h * 2))
    for i in range(0, w):
        img = out.subsurface((w - 1, 0, 1, h * 3)).convert_alpha()
        out.blit(img, (w + i, 0))
    for i in range(0, h):
        img = out.subsurface((0, h - 1, w * 3, 1)).convert_alpha()
        out.blit(img, (0, h + i))

    return out, w, h


todo = [
    ("button.normal", "dot.normal", None, 3, 3, "789456123"),
    ("button.hover", "dot.hover", None, 3, 3, "789456123"),
    ("button.down", "dot.down", None, 3, 3, "789456123"),
    ("checkbox.off.normal", "box.normal", None, 2, 2, "7913"),
    ("checkbox.on.normal", "box.down", "check", 2, 2, "7913"),
    ("checkbox.off.hover", "box.hover", None, 2, 2, "7913"),
    ("checkbox.on.hover", "box.hover", "check", 2, 2, "7913"),
    ("radio.off.normal", "dot.normal", None, 2, 2, "7913"),
    ("radio.on.normal", "dot.down", "radio", 2, 2, "7913"),
    ("radio.off.hover", "dot.hover", None, 2, 2, "7913"),
    ("radio.on.hover", "dot.hover", "radio", 2, 2, "7913"),
    ("tool.normal", "box.normal", None, 3, 3, "789456123"),
    ("tool.hover", "box.hover", None, 3, 3, "789456123"),
    ("tool.down", "box.down", None, 3, 3, "789456123"),
    ("hslider", "idot.normal", None, 3, 3, "789456123"),
    ("hslider.bar.normal", "dot.normal", None, 3, 3, "789456123"),
    ("hslider.bar.hover", "dot.hover", None, 3, 3, "789456123"),
    ("hslider.left", "sbox.normal", "left", 2, 2, "7913"),
    ("hslider.right", "sbox.normal", "right", 2, 2, "7913"),
    ("vslider", "idot.normal", None, 3, 3, "789456123"),
    ("vslider.bar.normal", "vdot.normal", None, 3, 3, "789456123"),
    ("vslider.bar.hover", "vdot.hover", None, 3, 3, "789456123"),
    ("vslider.up", "vsbox.normal", "up", 2, 2, "7913"),
    ("vslider.down", "vsbox.normal", "down", 2, 2, "7913"),
    ("dialog.close.normal", "rdot.hover", None, 2, 2, "7913"),
    ("dialog.close.hover", "rdot.hover", "x", 2, 2, "7913"),
    ("dialog.close.down", "rdot.down", "x", 2, 2, "7913"),
    ("menu.normal", "desktop", None, 1, 1, "7"),
    ("menu.hover", "box.normal", None, 3, 3, "789456123"),
    ("menu.down", "box.down", None, 3, 3, "789456123"),
    ("select.selected.normal", "box.normal", None, 3, 3, "788455122"),
    ("select.selected.hover", "box.hover", None, 3, 3, "788455122"),
    ("select.selected.down", "box.down", None, 3, 3, "788455122"),
    ("select.arrow.normal", "box.hover", None, 3, 3, "889556223"),
    ("select.arrow.hover", "box.hover", None, 3, 3, "889556223"),
    ("select.arrow.down", "box.down", None, 3, 3, "889556223"),
    ("progressbar", "sbox.normal", None, 3, 3, "789456123"),
    ("progressbar.bar", "box.hover", None, 3, 3, "789456123"),
]

for fname, img, over, ww, hh, s in todo:
    print(fname)
    img, w, h = prep(img)
    out = pygame.Surface((ww * w, hh * h), SWSURFACE | SRCALPHA, 32)
    out.fill((0, 0, 0, 0))
    n = 0
    for y in range(0, hh):
        for x in range(0, ww):
            c = int(s[n])
            xx, yy = (c - 1) % 3, 2 - (c - 1) / 3
            out.blit(img.subsurface((xx * w, yy * h, w, h)), (x * w, y * h))
            n += 1
    if over != None:
        over = pygame.image.load(over + ".png")
        out.blit(over, (0, 0))
    pygame.image.save(out, fname + ".tga")



if 0:
    try:
        # 0/0 #HACK: so i can CTRL-c out
        import psyco

        psyco.profile()
        print("psyco installed")
    except:
        print("psyco not installed")

import os, sys, time, copy


import pygame

# Import the android module. If we can't import it, set it to None - this
# lets us test it, and check to see if we want android-specific behavior.
try:
    import android
except ImportError:
    android = None


try:
    import pygame.mixer as mixer
except ImportError:
    import android_mixer

    pygame.mixer = android_mixer


from pygame.locals import *


from .pgu import engine
from .pgu import timer

from .const import *

if 0:
    # this can be used to figure out how big the desktop is...
    pygame.display.init()
    screen = pygame.display.set_mode((0, 0))
    SW, SH = screen.get_size()
    from . import const

    const.SW, const.SH = screen.get_size()


from . import states
from . import level
from . import units
from . import sounds

import traceback

from . import intro
from . import menu


class Game(engine.Game):
    def init(self):
        # self.timer = timer.Timer(FPS)
        self.timer = timer.Speedometer()
        self.clock = pygame.time.Clock()
        self.elapsed_time = 0
        self.last_time = time.time()
        self.cur_time = self.last_time

        self.clock.tick(FPS)

        # start a song playing here as things load.
        self.sm = sounds.SoundManager()
        # self.sm.PlayMusic( data_dir('music', "intro.ogg") )
        self.sm.PlayMusic(data_dir("intro", "intro1.ogg"))
        self.sm.Load()

        # for saving the states between levels.

        self.data_reset()
        self.should_restore_stats = True

    def tick(self):
        r = self.timer.tick()
        if r != None:
            # print r
            print(("fps :%s:" % self.clock.get_fps()))
        self.cur_ticks = self.clock.tick(FPS)  # to not limit it
        self.cur_time = time.time()

        self.elapsed_time = self.cur_time - self.last_time

        self.sm.Update(self.elapsed_time)
        # print "&"*20
        # print self.elapsed_time
        # print self.clock.get_fps()
        # for name in dir(self):
        #    print name,getattr(self,name)

    def save_castle_data(self):
        """ """

    def data_reset(self):
        self.data = {}
        self.data["levels"] = []
        self.upgradable_amounts = None
        self.castle_stats = None

        self.backup_upgradable_amounts = None
        self.backup_castle_stats = None

    def event(self, e):
        # Android-specific:
        if android:
            if android.check_pause():
                android.wait_for_resume()

        # capture special events on a top level,
        # should only be used for screen shots, forced quits,
        # magic cheat buttons, etc... debug keys and the like ...
        if e.type == QUIT:
            self.state = engine.Quit(self)
            return 1


#         if e.type == KEYDOWN and e.key == K_F10:
#             self.state = states.SPause(self,self.state)
#             return 1

#         if e.type == KEYDOWN and e.key == K_ESCAPE:
#             self.state = engine.Quit(self)
#             return 1


def do_main(no_intro=0, the_level=0):
    global flags

    print("hello from zanthor do_main")

    pygame.mixer.pre_init(22050, -16, 2, 1024)

    pygame.init()
    if pygame.mixer and not pygame.mixer.get_init():
        pygame.mixer = None
    pygame.font.init()
    # pygame.display.set_caption("The Wrath of ZANTHOR") #It's powered by steam!")
    pygame.display.set_caption(
        "The Wrath of ZANTHOR!  It's powered by steam!  h key for help"
    )

    # flags = FULLSCREEN
    print(flags)
    print((SW, SH))
    screen = pygame.display.set_mode((SW, SH), flags)

    # Map the back button to the escape key.
    if android:
        android.map_key(android.KEYCODE_BACK, pygame.K_ESCAPE)

    print("initialising joystick module")

    pygame.joystick.init()
    joystics = []
    num_joys = pygame.joystick.get_count()

    print(("initialising joys", num_joys))
    for x in range(num_joys):
        print(("x joy is:", x))
        j = pygame.joystick.Joystick(x)
        j.init()
        joystics.append(j)

    print("initialising joys: done.")

    game = Game()
    game.should_restore_stats = True
    game.backup_upgradable_amounts = None
    game.backup_castle_stats = None

    # the intro is not done first.
    # game.run(states.Title(game),screen)
    if no_intro == 0:
        game.run(intro.Intro(game), screen)

    elif no_intro == 1:
        # game.run(level.Level(game, 0,100),screen)
        game.run(menu.Menu(game), screen)
    elif no_intro == 3:
        # we load the level number from the levels defined in menu.
        n = menu.data[no_intro][4]
        perc = menu.data[no_intro][5]
        music = menu.data[no_intro][7]
        game.run(level.Level(game, n, perc, music), screen)
    else:
        game.run(level.Level(game, no_intro, 100, None), screen)


def main():
    global flags

    print("hello from zanthor: top of main")

    if "speed" in sys.argv:
        FPS = 65535

    the_level = None
    # flags ^= FULLSCREEN
    flags = 0
    if "fullscreen" in sys.argv or "full" in sys.argv:
        flags ^= FULLSCREEN

    if "nointro" in sys.argv or "no" in sys.argv:
        no_intro = 1
    elif "next" in sys.argv or "nextlevel" in sys.argv:
        no_intro = 2  # uhh, i removed this because i'm dumb
    elif "l" in sys.argv:
        # main.py l 5
        # we play the level given as an index into menu.data
        the_level = int(sys.argv[-1])
        no_intro = 3
    else:
        no_intro = 0

    # jump to any level by specifying, e.g. "level8.tga"
    for fname in sys.argv:
        if ".tga" in fname:
            no_intro = fname

    if "profile" in sys.argv:
        import hotshot
        import hotshot.stats
        import tempfile
        import os

        profile_data_fname = tempfile.mktemp("prf")
        try:
            prof = hotshot.Profile(profile_data_fname)
            prof.run("do_main(no_intro, the_level = the_level)")
            del prof
            s = hotshot.stats.load(profile_data_fname)
            s.strip_dirs()
            print("cumulative\n\n")
            s.sort_stats("cumulative").print_stats()
            print("By time.\n\n")
            s.sort_stats("time").print_stats()
            del s
        finally:
            # clean up the temporary file name.
            try:
                os.remove(profile_data_fname)
            except:
                # may have trouble deleting ;)
                pass
    else:
        try:
            do_main(no_intro, the_level=the_level)
        except:
            traceback.print_exc(sys.stderr)



"""All sliders and scroll bar widgets have the same parameters.
   
<pre>Slider(value,min,max,size)</pre> 
<dl>
<dt>value<dd>initial value
<dt>min<dd>minimum value
<dt>max<dd>maximum value
<dt>size<dd>size of bar in pixels
</dl>
"""
import pygame
from pygame.locals import *

from .const import *
from . import widget
from . import app
from . import table
from . import basic

_SLIDER_HORIZONTAL = 0
_SLIDER_VERTICAL = 1


class _slider(widget.Widget):
    def __init__(self, value, orient, min, max, size, step=1, **params):
        params.setdefault("cls", "slider")
        widget.Widget.__init__(self, **params)
        self.min, self.max, self.value, self.orient, self.size, self.step = (
            min,
            max,
            value,
            orient,
            size,
            step,
        )

    def paint(self, s):
        self.pcls = ""

        self.value = self.value

        if self.container.myhover is self:
            self.pcls = "hover"
        if self.container.myfocus is self and 1 in pygame.mouse.get_pressed():
            self.pcls = "down"

        r = pygame.rect.Rect(0, 0, self.rect.w, self.rect.h)
        if self.orient == _SLIDER_HORIZONTAL:
            r.x = (
                (self.value - self.min)
                * (self.rect.w - self.size)
                / (self.max - self.min)
            )
            r.w = self.size
        else:
            r.y = (
                (self.value - self.min)
                * (self.rect.h - self.size)
                / (self.max - self.min)
            )
            r.h = self.size

        self.bar = r

        app.App.app.theme.render(s, self.style.bar, r)

    def event(self, e):
        adj = 0
        if e.type == ENTER:
            self.repaint()
        elif e.type == EXIT:
            self.repaint()
        elif e.type == KEYDOWN:
            if e.key == K_TAB:
                next(self)
        elif e.type == MOUSEBUTTONDOWN:
            if self.bar.collidepoint(e.pos):
                self.grab = e.pos[0], e.pos[1]
                self.grab_value = self.value
            else:
                x, y, adj = e.pos[0], e.pos[1], 1
                self.grab = None
            self.repaint()
        elif e.type == MOUSEBUTTONUP:
            # x,y,adj = e.pos[0],e.pos[1],1
            self.repaint()
        elif e.type == MOUSEMOTION:
            if 1 in e.buttons and self.container.myfocus is self:
                if self.grab != None:
                    rel = e.pos[0] - self.grab[0], e.pos[1] - self.grab[1]
                    if self.orient == _SLIDER_HORIZONTAL:
                        d = self.rect.w - self.size
                        if d != 0:
                            self.value = self.grab_value + (
                                (self.max - self.min) * rel[0] / d
                            )
                    else:
                        d = self.rect.h - self.size
                        if d != 0:
                            self.value = self.grab_value + (
                                (self.max - self.min) * rel[1] / d
                            )
                else:
                    x, y, adj = e.pos[0], e.pos[1], 1

        if adj:
            if self.orient == _SLIDER_HORIZONTAL:
                d = self.size / 2 - (self.rect.w / (self.max - self.min + 1)) / 2
                self.value = (x - d) * (self.max - self.min) / (
                    self.rect.w - self.size + 1
                ) + self.min
            else:
                d = self.size / 2 - (self.rect.h / (self.max - self.min + 1)) / 2
                self.value = (y - d) * (self.max - self.min) / (
                    self.rect.h - self.size + 1
                ) + self.min

    def __setattr__(self, k, v):
        if k == "value":
            v = int(v)
            v = max(v, self.min)
            v = min(v, self.max)
        _v = self.__dict__.get(k, NOATTR)
        self.__dict__[k] = v
        if k == "value" and _v != NOATTR and _v != v:
            self.send(CHANGE)
            self.repaint()

        if hasattr(self, "size"):
            self.__dict__["size"] = max(
                self.size, min(self.style.width, self.style.height)
            )


class VSlider(_slider):
    """A verticle slider.

    <pre>VSlider(value,min,max,size)</pre>
    """

    def __init__(self, value, min, max, size, **params):
        params.setdefault("cls", "vslider")
        _slider.__init__(self, value, _SLIDER_VERTICAL, min, max, size, **params)


class HSlider(_slider):
    """A horizontal slider.

    <pre>HSlider(value,min,max,size)</pre>
    """

    def __init__(self, value, min, max, size, **params):
        params.setdefault("cls", "hslider")
        _slider.__init__(self, value, _SLIDER_HORIZONTAL, min, max, size, **params)


class xVScrollBar(_slider):
    def __init__(self, value, min, max, size, **params):
        params.setdefault("cls", "vscrollbar")
        _slider.__init__(self, value, _SLIDER_VERTICAL, min, max, size, **params)


class xHScrollBar(_slider):
    def __init__(self, value, min, max, size, **params):
        params.setdefault("cls", "hscrollbar")
        _slider.__init__(self, value, _SLIDER_HORIZONTAL, min, max, size, **params)


class HScrollBar(table.Table):
    """A horizontal scroll bar.

    <pre>HScrollBar(value,min,max,size,step=1)</pre>
    """

    def __init__(self, value, min, max, size, step=1, **params):
        params.setdefault("cls", "hscrollbar")

        table.Table.__init__(self, **params)

        self.slider = _slider(
            value,
            _SLIDER_HORIZONTAL,
            min,
            max,
            size,
            step=step,
            cls=self.cls + ".slider",
        )

        self.tr()

        self.minus = basic.Image(self.style.minus)
        self.minus.connect(MOUSEBUTTONDOWN, self._click, -1)
        self.td(self.minus)

        self.td(self.slider)
        self.slider.connect(CHANGE, self.send, CHANGE)

        self.minus = basic.Image(self.style.minus)
        self.minus.connect(MOUSEBUTTONDOWN, self._click, -1)
        self.td(self.minus)

        self.plus = basic.Image(self.style.plus)
        self.plus.connect(MOUSEBUTTONDOWN, self._click, 1)
        self.td(self.plus)

    def _click(self, value):
        self.slider.value += self.slider.step * value

    def resize(self, width=None, height=None):
        self.slider.style.width = self.style.width - (
            self.minus.style.width * 2 + self.plus.style.width
        )
        return table.Table.resize(self, width, height)

    def __setattr__(self, k, v):
        if k in ("min", "max", "value", "size", "step"):
            if k == "size":
                v = (
                    v
                    * (
                        self.style.width
                        - (self.minus.style.width * 2 + self.plus.style.width)
                    )
                    / self.style.width
                )
            return setattr(self.slider, k, v)
        self.__dict__[k] = v

    def __getattr__(self, k):
        if k in ("min", "max", "value", "step"):
            return getattr(self.slider, k)
        return table.Table.__getattr__(self, k)  # self.__dict__[k]


class VScrollBar(table.Table):
    """A vertical scroll bar.

    <pre>VScrollBar(value,min,max,size,step=1)</pre>
    """

    def __init__(self, value, min, max, size, step=1, **params):
        params.setdefault("cls", "vscrollbar")

        table.Table.__init__(self, **params)

        self.slider = _slider(
            value, _SLIDER_VERTICAL, min, max, size, step=step, cls=self.cls + ".slider"
        )

        self.tr()
        self.minus = basic.Image(self.style.minus)
        self.minus.connect(MOUSEBUTTONDOWN, self._click, -1)
        self.td(self.minus)

        self.tr()
        self.td(self.slider)
        self.slider.connect(CHANGE, self.send, CHANGE)

        self.tr()
        self.minus = basic.Image(self.style.minus)
        self.minus.connect(MOUSEBUTTONDOWN, self._click, -1)
        self.td(self.minus)

        self.tr()
        self.plus = basic.Image(self.style.plus)
        self.plus.connect(MOUSEBUTTONDOWN, self._click, 1)
        self.td(self.plus)

    def _click(self, value):
        self.slider.value += self.slider.step * value

    def resize(self, width=None, height=None):
        self.slider.style.height = self.style.height - (
            self.minus.style.height * 2 + self.plus.style.height
        )
        return table.Table.resize(self, width, height)

    def __setattr__(self, k, v):
        if k in ("min", "max", "value", "size", "step"):
            if k == "size":
                v = (
                    v
                    * (
                        self.style.height
                        - (self.minus.style.height * 2 + self.plus.style.height)
                    )
                    / self.style.height
                )
            return setattr(self.slider, k, v)
        self.__dict__[k] = v

    def __getattr__(self, k):
        if k in ("min", "max", "value", "step"):
            return getattr(self.slider, k)
        return table.Table.__getattr__(self, k)



"""a state engine. 
"""
import pygame
from pygame.locals import *


class State:
    """Template Class -- for a state.

    <pre>State(game,value...)</pre>

    <dl>
    <dt>game<dd>The state engine.
    <dt>value<dd>I usually pass in a custom value to a state
    </dl>

    <p>For all of the template methods, they should return None unless they return
    a new State to switch the engine to.</p>
    """

    def __init__(self, game, value=None):
        self.game, self.value = game, value

    def init(self):
        """Template Method - Initialize the state, called once the first time a state is selected.

        <pre>State.init()</pre>
        """
        return

    def paint(self, screen):
        """Template Method - Paint the screen.  Called once after the state is selected.

        <p>State is responsible for calling <tt>pygame.display.flip()</tt> or whatever.</p>

        <pre>State.paint(screen)</pre>
        """
        return

    def repaint(self):
        """Template Method - Request a repaint of this state.

        <pre>State.repaint()</pre>
        """
        self._paint = 1

    def update(self, screen):
        """Template Method - Update the screen.

        <p>State is responsible for calling <tt>pygame.display.update(updates)</tt> or whatever.</p>

        <pre>State.update(screen)</pre>
        """
        return

    def loop(self):
        """Template Method - Run a logic loop, called once per frame.

        <pre>State.loop()</pre>
        """
        return

    def event(self, e):
        """Template Method - Recieve an event.

        <pre>State.event(e)</pre>
        """
        return


class Quit(State):
    """A state to quit the state engine.

    <pre>Quit(game,value)</pre>
    """

    def init(self):
        self.game.quit = 1


class Game:
    """Template Class - The state engine."""

    def fnc(self, f, v=None):
        s = self.state
        if not hasattr(s, f):
            return 0
        f = getattr(s, f)
        if v != None:
            r = f(v)
        else:
            r = f()
        if r != None:
            self.state = r
            self.state._paint = 1
            return 1
        return 0

    def run(self, state, screen=None):
        """Run the state engine, this is a infinite loop (until a quit occurs).

        <pre>Game.run(state,screen=None)</pre>

        <dl>
        <dt>game<dd>a state engine
        <dt>screen<dd>the screen
        </dl>
        """
        self.quit = 0
        self.state = state
        if screen != None:
            self.screen = screen

        self.init()

        while not self.quit:
            self.loop()

    def loop(self):
        s = self.state
        if not hasattr(s, "_init") or s._init:
            s._init = 0
            if self.fnc("init"):
                return
        else:
            if self.fnc("loop"):
                return
        if not hasattr(s, "_paint") or s._paint:
            s._paint = 0
            if self.fnc("paint", self.screen):
                return
        else:
            if self.fnc("update", self.screen):
                return

        for e in pygame.event.get():
            # NOTE: this might break API?
            # if self.event(e): return
            if not self.event(e):
                if self.fnc("event", e):
                    return

        self.tick()
        return

    def init(self):
        """Template Method - called at the beginning of State.run() to initialize things.

        <pre>Game.init()</pre>
        """
        return

    def tick(self):
        """Template Method - called once per frame, usually for timer purposes.

        <pre>Game.tick()</pre>
        """
        pygame.time.wait(10)

    def event(self, e):
        """Template Method - called with each event, so the engine can capture special events.

        <pre>Game.event(e): return captured</pre>

        <p>return a True value if the event is captured and does not need to be passed onto the current
        state</p>
        """
        if e.type == QUIT:
            self.state = Quit(self)
            return 1


# vim: set filetype=python sts=4 sw=4 noet si :



"""
"""
import pygame
from pygame.locals import *

from . import container
from .const import *


class App(container.Container):
    """The top-level widget for an application.

    <pre>App(theme=None)</pre>

    <dl>
    <dt>theme<dd>an instance of a Theme, optional as it will use the default Theme class.
    </dl>

    <strong>Basic Example</strong>
    <code>
    app = gui.App()
    app.run(widget=widget,screen=screen)
    </code>

    <strong>Integrated Example</strong>
    <code>
    app = gui.App()
    gui.init(widget=widget)
    while 1:
        for e in pygame.event.get():
            app.event(e)
        app.update(screen)
    </code>



    """

    def __init__(self, theme=None, **params):
        App.app = self

        if theme == None:
            from .theme import Theme

            theme = Theme()
        self.theme = theme

        params["decorate"] = "app"
        container.Container.__init__(self, **params)
        self._quit = False
        self.widget = None
        self._chsize = False
        self._repaint = False

        self.screen = None
        self.container = None
        self.events = []

    def resize(self):
        screen = self.screen
        w = self.widget
        wsize = 0

        # 5 cases

        # input screen is already set use its size
        if screen:
            self.screen = screen
            width, height = screen.get_width(), screen.get_height()

        # display.screen
        elif pygame.display.get_surface():
            screen = pygame.display.get_surface()
            self.screen = screen
            width, height = screen.get_width(), screen.get_height()

        # app has width,height
        elif self.style.width != 0 and self.style.height != 0:
            screen = pygame.display.set_mode(
                (self.style.width, self.style.height), SWSURFACE
            )
            self.screen = screen
            width, height = screen.get_width(), screen.get_height()

        # widget has width,height, or its own size..
        else:
            wsize = 1
            width, height = w.rect.w, w.rect.h = w.resize()
            # w._resize()
            screen = pygame.display.set_mode((width, height), SWSURFACE)
            self.screen = screen

        # use screen to set up size of this widget
        self.style.width, self.style.height = width, height
        self.rect.w, self.rect.h = width, height
        self.rect.x, self.rect.y = 0, 0

        w.rect.x, w.rect.y = 0, 0
        w.rect.w, w.rect.h = w.resize(width, height)

        for w in self.windows:
            w.rect.w, w.rect.h = w.resize()

    def init(
        self, widget=None, screen=None
    ):  # TODO widget= could conflict with module widget
        """Initialize the application.

        <pre>App.init(widget=None,screen=None)</pre>

        <dl>
        <dt>widget<dd>main widget
        <dt>screen<dd>pygame.Surface to render to
        </dl>
        """

        App.app = self

        if widget:
            self.widget = widget
        if screen:
            self.screen = screen

        self.resize()

        w = self.widget

        self.widgets = []
        self.widgets.append(w)
        w.container = self
        self.focus(w)

        pygame.key.set_repeat(500, 30)

        self._repaint = True
        self._quit = False

        self.send(INIT)

    def event(self, e):
        """Pass an event to the main widget.

        <pre>App.event(e)</pre>

        <dl>
        <dt>e<dd>event
        </dl>
        """
        App.app = self
        # NOTE: might want to deal with ACTIVEEVENT in the future.
        self.send(e.type, e)
        container.Container.event(self, e)
        if e.type == MOUSEBUTTONUP:
            sub = pygame.event.Event(CLICK, {"button": e.button, "pos": e.pos})
            self.send(sub.type, sub)
            container.Container.event(self, sub)

    def loop(self):
        App.app = self
        s = self.screen
        for e in pygame.event.get():
            if not (e.type == QUIT and self.mywindow):
                self.event(e)
        us = self.update(s)
        pygame.display.update(us)

    def update(self, screen):
        """Update the screen.

        <dl>
        <dt>screen<dd>pygame surface
        </dl>
        """
        if self._chsize:
            self.resize()
            self._chsize = False
        if self._repaint:
            if hasattr(self, "background"):
                self.background.paint(screen)
            self.paint(screen)
            self._repaint = False
            return [pygame.Rect(0, 0, screen.get_width(), screen.get_height())]
        else:
            us = container.Container.update(self, screen)
            return us

    def run(self, widget=None, screen=None):
        """Run an application.

        <p>Automatically calls <tt>App.init</tt> and then forever loops <tt>App.event</tt> and <tt>App.update</tt></p>

        <dl>
        <dt>widget<dd>main widget
        <dt>screen<dd>pygame.Surface to render to
        </dl>
        """
        self.init(widget, screen)
        while not self._quit:
            self.loop()
            pygame.time.wait(10)

    def reupdate(self, w=None):
        pass

    def repaint(self, w=None):
        self._repaint = True

    def repaintall(self):
        self._repaint = True

    def chsize(self):
        self._chsize = True
        self._repaint = True

    def quit(self, value=None):
        self._quit = True


class Desktop(App):
    """Create an App using the <tt>desktop</tt> theme class.

    <pre>Desktop()</pre>
    """

    def __init__(self, **params):
        params.setdefault("cls", "desktop")
        App.__init__(self, **params)



from . import isovid
from . import algo

from .const import *
from .units import CoalTruck


def truck_new(g, t, v):
    print("trucks disabled")
    return
    g.clayer[t.ty][t.tx] = 0
    s = isovid.Sprite(g.images["truck"], t.rect)

    s.frame = 0

    g.sprites.append(s)
    s.loop = truck_loop
    s.hit = truck_hit
    s.groups = g.string2groups("truck")
    s.agroups = g.string2groups("castle")

    s.origin = t.tx, t.ty
    s.state = "init"
    s.unit = CoalTruck()


def dist(a, b):
    return abs(a[0] - b[0]) + abs(a[1] - b[1])


def sign(v):
    if v == 0:
        return v
    return v / abs(v)


def truck_loop(g, s):
    s.unit.loop()

    s.frame += 1

    tw, th = g.iso_w, g.iso_h
    ww, hh = g.size
    sx, sy = s.rect.centerx / tw, s.rect.centery / th
    if s.state == "init":
        best_s, best_d = None, 65535
        for ss in g.sprites:
            if hasattr(ss, "type") and ss.type == "factory":
                fx, fy = ss.rect.centerx / tw, ss.rect.centery / th
                dst = dist((sx, sy), (fx, fy))
                if dst < best_d:
                    best_s, best_d = ss, dst
        s.factory = best_s
        s.state = "search"
        return

    if s.state == "search":
        coals = []
        for y in range(0, ww):
            for x in range(0, hh):
                if g.tlayer[y][x] == 1:  # HACK: if is coal
                    coals.append((x, y))
        if len(coals) == 0:
            s.state = "wait"
            return

        best, best_d = None, 65535
        fx, fy = s.factory.rect.centerx / tw, s.factory.rect.centery / th
        for cx, cy in coals:
            dst = dist((cx, cy), (fx, fy))
            if dst < best_d:
                best, best_d = (cx, cy), dst

        cx, cy = best
        s.target = cx, cy
        s.path = algo.astar((sx, sy), (cx, cy), g.truck_layer, dist)
        s.state = "coal"

    # once a second check if some coal has appeared...
    if s.state == "wait":
        if s.frame % FPS == 0:
            s.state = "search"
        return

    if s.state == "coal":
        fx, fy = s.target
        n = g.tlayer[fy][fx]
        if n != 1:  # HACK: coal?
            s.state = "search"
            return
        if (sx, sy) == s.target:
            s.coal = 1
            g.set((sx, sy), 0)

            fx, fy = s.factory.rect.centerx / tw, s.factory.rect.centery / th
            s.target = fx, fy
            s.path = algo.astar((sx, sy), (fx, fy), g.truck_layer, dist)
            s.state = "factory"

    if s.state == "factory":
        if (sx, sy) == s.target:
            s.coal = 0
            # NOTE: give the coal to the factory here....
            s.state = "search"
            return

    path = s.path
    while len(path) > 1:
        bx, by = path[1]
        if (bx, by) == (sx, sy):
            path.pop(0)
            continue
        dx, dy = bx - sx, by - sy
        v = 4
        s.rect.x += sign(dx) * v
        s.rect.y += sign(dy) * v
        break


def truck_hit(g, a, b):
    pass



from . import isovid

from .const import *

# from pgu import algo
from . import algo

from .units import Robot

from . import sound_info


import random


MIN = 8.0
MIN_RAND = 4
MAX = 16.0
MAX_RAND = 8
RADIUS = 32.0
RADIUS_RAND = 12
CENTER = 4.0
CENTER_RAND = 2
SEARCH = 128.0


def robot_new(g, t, v):
    g.clayer[t.ty][t.tx] = 0
    s = isovid.Sprite(g.images["robot"], t.rect)

    s.frame = 0
    s.origin = t.tx, t.ty

    g.sprites.append(s)
    s.loop = robot_loop
    # if random.randrange(0,20) == 0:
    #    s.loop = robot_loop_2
    s.hit = robot_hit
    s.groups = g.string2groups("robot")
    s.agroups = 0  # g.string2groups("castle")
    s.path = []
    s.state = "alive"
    s.type = "robot"

    s.unit = Robot()

    s.x, s.y = float(s.rect.x), float(s.rect.y)
    s._x, s._y = s.x, s.y
    s._min = s.min = MIN + random.randrange(-MIN_RAND, MIN_RAND)
    s.max = MAX + random.randrange(-MAX_RAND, MAX_RAND)
    s.radius = RADIUS + random.randrange(-RADIUS_RAND, RADIUS_RAND)
    s.center = CENTER + random.randrange(-CENTER_RAND, CENTER_RAND)
    s.search = SEARCH
    s._flock_frame = 0
    s.near = []
    s.avx = 0
    s.avy = 0
    s.ax = 0
    s.ay = 0

    g.robots.append(s)

    ty, tx = t.ty, t.tx
    g.robot_list_layer[ty][tx].append(s)


def dist(a, b):
    return abs(a[0] - b[0]) + abs(a[1] - b[1])


def sign(v):
    if v == 0:
        return v
    return v // abs(v)


def sdist(a, b):
    return abs(a.rect.x - b.rect.x) + abs(a.rect.y - b.rect.y)


def robot_shove(g, s, xxx_todo_changeme):
    (x, y) = xxx_todo_changeme
    tw, th = g.iso_w, g.iso_h
    r = pygame.Rect(s.rect)
    dx, dy = x - r.x, y - r.y
    dist = (dx * dx + dy * dy) ** 0.5
    dx, dy = dx // dist, dy // dist
    n = 0
    inc = 8.0
    xx, yy = r.x, r.y
    while n < dist:
        # play them in order.  if there is one already playing, don't play it.
        g.level.game.sm.Play(sound_info.peasants.nextone(), wait=2)
        p = pygame.Rect(r)
        n += inc  # sure, why not 8!
        xx += dx * inc
        yy += dy * inc
        r.x, r.y = xx, yy
        tx, ty = r.centerx // tw, r.centery // th
        if g.robot_layer[ty][tx]:
            s.x, s.y = p.x, p.y
            return
    s.x, s.y = x, y


def robot_loop(g, s):
    if s.state == "dead":
        if s in g.sprites:
            g.sprites.remove(s)
        if s in g.robots:
            g.robots.remove(s)
        return

    tw, th = g.iso_w, g.iso_h
    s.min = max(s._min, min(s.max, s.min - 0.5))

    s.rect.x, s.rect.y = s.x, s.y
    tx, ty = s.rect.centerx // tw, s.rect.centery // th

    if g.robot_layer[ty][tx]:
        s.rect.x, s.rect.y = s._x, s._y
        tx, ty = s.rect.centerx // tw, s.rect.centery // th
        if g.robot_layer[ty][tx]:
            # py3k?
            # x, y = s.px, s.py
            x, y = tx, ty
            s.rect.x, s.rect.y = x, y
        s.x, s.y = s.rect.x, s.rect.y

    tx, ty = s.rect.centerx // tw, s.rect.centery // th
    if g.robot_layer[ty][tx]:
        if s in g.sprites:
            g.sprites.remove(s)
        if s in g.robots:
            g.robots.remove(s)


def robot_init(g, s):
    rdist = []
    for r in g.robots:
        rdist.append((sdist(r, s), r))
    rdist.sort()
    s.robots = []
    for d, r in rdist[0:3]:
        s.robots.append(r)


def robot_loop_bilge(g, s):
    print((len(g.robots)))
    return
    # s.unit.loop()

    # if random.randrange(0,30) == 0:
    #    s.init = True

    # if s.init:
    #    robot_init(g,s)
    #    s.init = False

    tw, th = g.iso_w, g.iso_h
    tx, ty = s.rect.centerx // tw, s.rect.centery // th

    g.robot_list_layer[ty][tx].remove(s)

    xx, yy = 0, 0
    total = 0.0
    #     for r in s.robots:
    for _ty in range(ty - 1, ty + 2):
        for _tx in range(tx - 1, tx + 2):
            # for _tx,_ty in [(tx-1,ty),(tx+1,ty),(tx,ty-1),(tx,ty+1)]:
            for r in g.robot_list_layer[_ty][_tx]:
                sx, sy = s.frect.x, s.frect.y
                rx, ry = r.frect.x, r.frect.y
                px, py = r._frect.x, r._frect.y

                # go towards neighbor
                f = 1.0
                xx, yy, total = xx + rx * f, yy + ry * f, total + f

                # stear away from neighbor
                dx, dy = sx - rx, sy - ry
                while dx == 0 and dy == 0:
                    dx, dy = random.randrange(-1, 2), random.randrange(-1, 2)
                dist = abs(dx) + abs(dy)
                sep = 32
                x, y = rx + (sep * dx // dist), ry + (sep * dy // dist)
                f = 1.0
                xx, yy, total = xx + x * f, yy + y * f, total + f

                # stear towards you + neighbors vector
                dx, dy = rx - px, ry - py
                x, y = sx + dx, sy + dy
                f = 1.0
                xx, yy, total = xx + x * f, yy + y * f, total + f

    if total != 0:
        xx = int(xx // total)
        yy = int(yy // total)
    else:
        xx, yy = s.rect.x, s.rect.y

    px, py = s.rect.x, s.rect.y

    dx, dy = xx - px, yy - py
    mx = 16
    val = abs(dx) + abs(dy)
    if val > mx:
        xx = px + dx * mx // val
        yy = py + dy * mx // val

    if val == 0:
        dx, dy = s.frect.x - s._frect.x, s.frect.y - s._frect.y
        xx, yy = px + dx, py + dy

    ty, tx = xx // tw, yy // th
    if g.robot_layer[ty][tx]:
        xx, yy = s.rect.x, s.rect.y

    s.rect.x, s.rect.y = xx, yy

    tx, ty = s.rect.centerx // tw, s.rect.centery // th
    g.robot_list_layer[ty][tx].append(s)


def robot_loop_2(g, s):
    tw, th = g.iso_w, g.iso_h
    sx, sy = s.rect.centerx // tw, s.rect.centery // th

    if s.frame % FPS in (0, 10, 20, 30):  # HACK, uh oh!
        # rint g.slayer#

        # sx,sy = s.rect.centerx/tw,s.rect.centery/th
        p = g.castle
        px, py = p.rect.centerx // tw, p.rect.centery // th
        ox, oy = s.origin[0], s.origin[1]
        dx, dy = px - ox, py - oy  # px-sx,py-sy

        if (dx * dx + dy * dy) ** 0.5 < 12:  # follow if near by origin
            sx, sy = s.rect.centerx // tw, s.rect.centery // th
            p = g.castle
            px, py = p.rect.centerx // tw, p.rect.centery // th
            s.path = algo.astar((sx, sy), (px, py), g.robot_layer, dist)
        else:  # else return to origin
            s.path = algo.astar((sx, sy), (ox, oy), g.robot_layer, dist)

    path = s.path
    if 1:  # or len(path) < 12:
        while len(path) > 1:
            bx, by = path[1]
            if (bx, by) == (sx, sy):
                path.pop(0)
                continue
            dx, dy = bx - sx, by - sy
            v = 4
            s.rect.x += sign(dx) * v
            s.rect.y += sign(dy) * v
            break

    s.frame += 1


def robot_hit(g, a, b):
    pass



from . import isovid

# import pygame


# set f < 1 for infinite
def effect_new(g, rect, e, f):
    s = isovid.Sprite(g.images["blank"], rect)
    s.effect = e

    s.frames = f
    g.sprites.append(s)

    s.loop = effect_loop

    return s


def effect_loop(g, s):
    s.frames -= 1
    if s.frames == 0:
        g.sprites.remove(s)



"""
"""
from .const import *
from . import table
from . import basic, button


class _Menu_Options(table.Table):
    def __init__(self, menu, **params):
        table.Table.__init__(self, **params)

        self.menu = menu

    def event(self, e):
        handled = False
        if e.type == MOUSEMOTION:
            abspos = e.pos[0] + self.rect.x, e.pos[1] + self.rect.y
            for w in self.menu.container.widgets:
                if not w is self.menu:
                    if w.rect.collidepoint(abspos):
                        self.menu._close(None)
                        w._open(None)
                        handled = True
        elif e.type == MOUSEBUTTONUP:
            abspos = e.pos[0] + self.rect.x, e.pos[1] + self.rect.y
            if not self.rect.collidepoint(
                abspos
            ) and not self.menu.container.rect.collidepoint(abspos):
                self.close()
                handled = True

        if not handled:
            table.Table.event(self, e)


class _Menu(button.Button):
    def __init__(
        self, widget=None, **params
    ):  # TODO widget= could conflict with module widget
        params.setdefault("cls", "menu")
        button.Button.__init__(self, widget, **params)

        self._cls = self.cls
        self.options = _Menu_Options(self, cls=self.cls + ".options")

        # self.options.connect(BLUR,self._close,None)
        self.connect(CLICK, self._open, None)

        self.pos = 0

    def _open(self, value):
        self.cls = self._cls + "-open"
        self.repaint()
        self.container.open(self.options, self.rect.x, self.rect.bottom)
        # action_open({'container':self.container,'window':self.options,'x':self.rect.x,'y':self.rect.bottom})
        self.options.blur(self.options.myfocus)
        self.options.focus()
        self.repaint()

    def _close(self, value):
        self.cls = self._cls
        self.repaint()
        self.options.close()

    def _value(self, value):
        # self.options.close()
        self._close(None)
        if value["fnc"] != None:
            value["fnc"](value["value"])

    #    def resize(self,width=None,height=None):
    #        w,h = button.Button.resize(self,width,height)
    #
    #        return w,h
    #        #self.options._resize()

    def add(self, w, fnc=None, value=None):
        # w.resize()

        w.style.align = -1
        b = button.Button(w, cls=self.cls + ".option")
        # b.style.hexpand = 1
        b.connect(CLICK, self._value, {"fnc": fnc, "value": value})

        # self.options.add(b,0,self.pos,1,1,-1,0)
        self.options.tr()
        self.options.add(b)
        # self.pos += 1

        return b


class Menus(table.Table):
    """A drop down menu bar.

    <pre>Menus(data)</pre>

    <dl>
    <dt>data<dd>Menu data, a list of (path,fnc,value), see example below
    </dl>

    <strong>Example</strong>
    <code>
    data = [
        ('File/Save',fnc_save,None),
        ('File/New',fnc_new,None),
        ('Edit/Copy',fnc_copy,None),
        ('Edit/Cut',fnc_cut,None),
        ('Help/About',fnc_help,help_about_content),
        ('Help/Reference',fnc_help,help_reference_content),
        ]
    w = Menus(data)
    """

    def __init__(self, data, menu_cls="menu", **params):
        params.setdefault("cls", "menus")
        table.Table.__init__(self, **params)

        n, m, mt = 0, None, None
        for path, cmd, value in data:
            parts = path.split("/")
            if parts[0] != mt:
                mt = parts[0]
                m = _Menu(basic.Label(mt, cls=menu_cls + ".label"), cls=menu_cls)
                self.add(m, n, 0)
                n += 1
            m.add(basic.Label(parts[1], cls=m.cls + ".option.label"), cmd, value)
            # m.resize()



from . import isovid
from .units import Factory


def factory_new(g, t, value):
    print(("factory", g.clayer[t.ty][t.tx], t.rect))

    g.clayer[t.ty][t.tx] = 0
    g.blayer[t.ty][t.tx] = 0x14
    g.blayer[t.ty - 1][t.tx] = 0x14
    g.blayer[t.ty][t.tx - 1] = 0x14
    g.blayer[t.ty + 1][t.tx] = 0x14
    g.blayer[t.ty][t.tx + 1] = 0x14

    # s = isovid.Sprite(g.images['factory.3'],(0,0))
    # s.rect.centerx,s.rect.centery = t.rect.centerx,t.rect.centery/2
    s = isovid.Sprite(g.images["factory.3"], t.rect)
    g.sprites.append(s)
    print((s.rect))
    tx, ty = s.rect.centerx / g.iso_w, s.rect.centery / g.iso_h
    print((tx, ty))

    # NOTE: this is just to show off!
    # import steam
    # s.effect = steam.Effect(32,1)

    from . import fire

    s.effect = fire.Effect(16, 1)

    # s.groups = g.string2groups('coal')
    # s.agroups = g.string2groups('castle')
    # s.hit = coal_hit

    # s.groups = g.string2groups('factory')
    # s.agroups =g.string2groups("castle")
    s.hit = factory_hit

    s.loop = factory_loop
    s.frame = 0
    s.type = "factory"

    s.unit = Factory()


def factory_loop(g, s):
    s.unit.loop()

    tx, ty = (
        s.rect.centerx / 32,
        s.rect.centery / 32,
    )  # g.view_to_iso((s.rect.centerx,s.rect.centery))
    # print s.rect
    # print tx,ty
    # print g.tlayer[ty][tx]
    s.frame += 1


def factory_hit(g, a, b):
    pass



import os, random, time

import pygame

# from pgu import isovid
from . import isovid
from .pgu import gui

from pygame.locals import *
from .const import *
from . import const
from . import human, castle, items
from . import interface
# from . import states

from . import truck, cannon, robot
from . import factory


from . import tiles
from . import menu
from . import flock


class Level:
    def __init__(self, game, _round, perc, music):
        self.game = game
        self.round = _round
        self.percent = perc

        self.tv = tv = isovid.Isovid()
        self.tv.level = self

        self.last_mouse_event_time = time.time()
        self.last_mouse_moves = 0

        if music:
            if pygame.mixer:
                pygame.mixer.music.load(data_dir("intro", music))
                pygame.mixer.music.play(-1)

        tdata = {
            0x01: ("castle", tiles.tile_coal, None),
            0x02: ("castle", tiles.tile_water, None),
            # 0x03:('castle',tiles.tile_block,{'top':1,'bottom':1,'left':1,'right':1}),
            # 0x04:('castle',tiles.tile_block,{'top':1,'bottom':1,'left':1,'right':1}),
            # 0x05:('castle',tiles.tile_block,{'top':1,'bottom':1,'left':1,'right':1}),
            # 0x06:('castle',tiles.tile_block,{'top':1,'bottom':1,'left':1,'right':1}),
            # 0x03:('cball',tiles.tile_wall,None),
            0x04: ("cball", tiles.tile_wall, None),
            0x05: ("cball", tiles.tile_wall, None),
            0x06: ("cball", tiles.tile_wall, None),
            0x07: ("castle", tiles.tile_rubble, None),
            8: ("castle", tiles.tile_part, None),
            9: ("castle", tiles.tile_part, None),
            10: ("castle", tiles.tile_part, None),
            11: ("castle", tiles.tile_part, None),
            12: ("castle", tiles.tile_part, None),
            13: ("castle", tiles.tile_part, None),
            14: ("castle", tiles.tile_part, None),
            15: ("castle", tiles.tile_part, None),
            0x1C: (
                "castle,cball,robot",
                tiles.tile_limit,
                None,
            ),  # remove item from level if it goes here.
        }
        # TODO: FIXME:

        tv.tga_load_tiles(data_dir("gfx", "tiles2.tga"), (32, 64), tdata)

        tv.tiles[1].item = items.Coal(3.0)  # HERE!!
        tv.tiles[2].item = items.Water(3.0)

        tv.tiles[8].item = items.Part("UpEngine Efficiency", "")
        tv.tiles[9].item = items.Part("UpEngine Speed", "")
        tv.tiles[10].item = items.Part("UpCannon Balls", "")
        tv.tiles[11].item = items.Part("UpArmour", "")
        tv.tiles[12].item = items.Part("UpSteam Tank", "")
        tv.tiles[13].item = items.Part("UpWater Tank", "")
        tv.tiles[14].item = items.Part("UpCoal Tank", "")
        tv.tiles[15].item = items.Part("UpCannon Power", "")

        tv.tiles[7].item = items.Rubble()

        # TODO: add in the rubble item.

        fname = self.round
        if fname == 0:
            fname = "test2.tga"
        tv.tga_load_level(data_dir("levels", fname), 1)

        # NOTE: blur the background
        for n in range(0, 1):
            layer = tv.blayer
            nlayer = [[0 for x in range(0, tv.size[0])] for y in range(0, tv.size[1])]
            w, h = tv.size
            for y in range(0, h):
                for x in range(0, w):
                    v = 0
                    for dx, dy in [
                        (-1, -1),
                        (0, -1),
                        (1, -1),
                        (-1, 0),
                        (1, 0),
                        (-1, 1),
                        (0, 1),
                        (1, 1),
                    ]:
                        xx = min(max(x + dx, 0), w - 1)
                        yy = min(max(y + dy, 0), h - 1)
                        v += layer[yy][xx]
                    v /= 8
                    nlayer[y][x] = v
            for y in range(0, h):
                for x in range(0, w):
                    layer[y][x] = nlayer[y][x]

        # NOTE: this layer is for enemies "astar" stuff
        tv.truck_layer = [
            [0 for x in range(0, tv.size[0])] for y in range(0, tv.size[1])
        ]
        tv.robot_layer = [
            [0 for x in range(0, tv.size[0])] for y in range(0, tv.size[1])
        ]
        tv.castle_layer = [
            [0 for x in range(0, tv.size[0])] for y in range(0, tv.size[1])
        ]
        # this is for robots to see who is near by
        tv.robot_list_layer = [
            [[] for x in range(0, tv.size[0])] for y in range(0, tv.size[1])
        ]

        # this is to init all the astar layers, forcefully...
        # since load_level shortcuts all the .set() stuff
        for y in range(0, tv.size[1]):
            for x in range(0, tv.size[0]):
                tv.set((x, y), tv.get((x, y)))

        tv.view.x = -220
        tv.view.y = -64
        tv.view.w, tv.view.h = S_VIEW.w, S_VIEW.h
        # tv.view.w,tv.view.h = SW,SH
        # print "view"
        # print tv.view

        tv.tile_w, tv.tile_h = 32, 64
        tv.iso_w, tv.iso_h = 32, 32
        tv.base_w, tv.base_h = 32, 16
        tv.iso_z = 1

        # best#shape = (8,64-24,16,16)
        shape = (0, 48, 32, 16)

        # shape = (2,64-30,28,28)
        # shape = (16,64-16,1,1)

        idata = [
            # ('castle',data_dir('gfx', 'castle1.png'),(8,64-16,16,8)),
            # ('coal',data_dir('gfx', 'coal.png'),(8,64-16,16,8)),
            # ('castle',tv.tiles[8].image,shape),
            ("coal", tv.tiles[1].image, shape),
            ("water", tv.tiles[2].image, shape),
            ("UpEngine Efficiency", tv.tiles[8].image, shape),
            ("UpEngine Speed", tv.tiles[9].image, shape),
            ("UpCannon Balls", tv.tiles[10].image, shape),
            ("UpArmour", tv.tiles[11].image, shape),
            ("UpSteam Tank", tv.tiles[12].image, shape),
            ("UpWater Tank", tv.tiles[13].image, shape),
            ("UpCoal Tank", tv.tiles[14].image, shape),
            ("UpCannon Power", tv.tiles[15].image, shape),
            ("factory.1", tv.tiles[3].image, shape),
            ("factory.2", tv.tiles[4].image, shape),
            ("factory.3", tv.tiles[5].image, shape),
            ("cannon_tower", tv.tiles[11].image, shape),
            ("wall", tv.tiles[23].image, shape),
            ("human", tv.tiles[9].image, shape),
            ("baddie_robot", tv.tiles[10].image, shape),
            ("truck", tv.tiles[15].image, shape),
            ("cannon", tv.tiles[11].image, shape),
            # ('cball',tv.tiles[14].image,shape),
            ("cball", tv.tiles[32].image, shape),
            ("cball2", tv.tiles[33].image, shape),
            ("cball3", tv.tiles[34].image, shape),
            ("cball4", tv.tiles[35].image, shape),
            (
                "robot",
                tv.tiles[12].image.subsurface(0, 24, 32, 32 + 8),
                (8, 24, 16, 16),
            ),
        ]
        try:
            timg = pygame.image.load(data_dir("gfx", "tiles2.tga")).convert_alpha()
        except:
            timg = pygame.image.load(data_dir("gfx", "tiles2.png")).convert_alpha()
        img = timg.subsurface((0, 64 * 2 + 32, 64, 64 + 32))
        shape = (16, 32 + 32, 32, 32)
        idata.append(("castle", img, shape))
        idata.append(("castle.left", img, shape))
        idata.append(("castle.right", pygame.transform.flip(img, -1, 0), shape))

        cdata = {
            1: (castle.castle_new, None),
            # 2:(coal_new,None),
            # 3:(water_new,None),
            4: (factory.factory_new, None),
            # 16:(cannon_tower_new,None),
            # 17:(baddie_robot_new,None),
            # 18:(wall_new,None),
            # 19:(human_new,None),
            5: (truck.truck_new, None),
            6: (cannon.cannon_new, None),
            7: (robot.robot_new, None),
        }

        # tv.load_images(idata) #should do this later, but i'm
        # using a quick hack for now
        for name, img, shape in idata:
            tv.images[name] = img, shape

        tw, th = tv.iso_w, tv.iso_h
        w, h = tv.size
        border = tw * 2
        rect = pygame.Rect(-border, -border, tw * w + border * 2, th * h + border * 2)
        # tv.robots = flock.Flock(rect,tw*3/2)
        tv.robots = flock.Flock(rect, tw * 2)
        tv.run_codes(cdata, (0, 0, tv.size[0], tv.size[1]))

        tv.robots_total = len(tv.robots)

        #
        self.interface = interface.Interface()
        self.interface.load()

        self.interface.tv = tv
        self.interface.max_robots = len(tv.robots)

        # ??? what does this code do???

        for k, v in list(tv.images.items()):
            shape = None
            print(v)
            if len(v) > 1:
                img, shape = v
            else:
                img = v

            img2 = pygame.Surface((img.get_width(), img.get_height())).convert()
            ck = (255, 0, 255)
            img2.fill(ck)
            img2.blit(img, (0, 0))
            img2.set_colorkey(ck)

            if shape != None:
                tv.images[k] = img2, shape
            else:
                tv.images[k] = img2

        n = 0
        for t in tv.tiles:
            if t != None and t.image != None:
                ck = (255, 0, 255, 255)
                img = t.image
                minh = 64
                maxh = 0
                x = 16
                for y in range(0, 64):
                    c = img.get_at((x, y))
                    if c[3] == 255:
                        minh = min(minh, y)
                        maxh = max(maxh, y)

                if minh > maxh:
                    img2 = None
                    print(("ahh", n))
                elif minh != 0 or maxh != 63:
                    t.h = maxh - minh
                    img = img.subsurface(0, minh, 32, maxh - minh)
                    img2 = pygame.Surface((img.get_width(), img.get_height())).convert()
                    ck = (255, 0, 255)
                    img2.fill(ck)
                    img2.blit(img, (0, 0))
                    img2.set_colorkey(ck)
                else:
                    img2 = pygame.Surface((img.get_width(), img.get_height())).convert()
                    ck = (255, 0, 255)
                    img2.fill(ck)
                    img2.blit(img, (0, 0))
                    img2.set_colorkey(ck)

                t.image = img2
            n += 1

        # load a few more images...
        tv.images["hole"] = pygame.image.load(
            data_dir("gfx", "hole.png")
        ).convert_alpha()
        tv.images["hole2"] = pygame.image.load(
            data_dir("gfx", "hole2.png")
        ).convert_alpha()
        tv.images["hole3"] = pygame.image.load(
            data_dir("gfx", "hole3.png")
        ).convert_alpha()
        tv.images["hole4"] = pygame.image.load(
            data_dir("gfx", "hole4.png")
        ).convert_alpha()

        img = pygame.Surface((1, 1)).convert()
        img.fill((255, 0, 255))
        img.set_colorkey((255, 0, 255))
        tv.images["blank"] = img

        tv.auto_scroll = True

        try:
            self.hairs = pygame.image.load(data_dir("gfx", "hairs.tga")).convert_alpha()
        except:
            self.hairs = pygame.image.load(data_dir("gfx", "hairs.png")).convert_alpha()

        self.frames = 0
        self.hair_update = None

        self.game.magic_castle = tv.castle
        self.interface.stats_draw.robots_total = 0
        self.interface.stats_draw.robots_update = 0

        self.last_show_hairs = False

    def init(self):
        # blah blah
        pass

    def paint(self, screen):
        tv = self.tv

        # we set these variables because the can change.
        SCROLL_MOUSE, SCROLL_AUTO, SCROLL_BORDER = const.get_mouse_info()

        show_hairs = True

        if self.frames > FPS:  # don't scroll for at least a second...
            sv = pygame.Rect(S_VIEW)
            border = 8
            sv.x += border
            sv.y += border
            sv.w -= border * 2
            sv.h -= border * 2
            v = tv.view
            x, y = pygame.mouse.get_pos()
            inc = SCROLL_MOUSE
            if not sv.collidepoint((x, y)):
                tv.auto_scroll = False
            if x <= sv.left:
                v.x -= inc
                show_hairs = False
            if x >= sv.right:
                v.x += inc
                show_hairs = False
            if y <= sv.top:
                v.y -= inc
                show_hairs = False
            if y >= sv.bottom:
                v.y += inc
                show_hairs = False
            c = tv.castle.irect
            if not v.contains(c):
                if v.left > c.left:
                    v.left = c.left
                if v.right < c.right:
                    v.right = c.right
                if v.top > c.top:
                    v.top = c.top
                if v.bottom < c.bottom:
                    v.bottom = c.bottom

        v = tv.view
        p = tv.castle.irect
        border = SCROLL_BORDER

        inc = SCROLL_AUTO
        if self.frames < 5:  # HACK, i don't know what it'll be...
            inc = 16384

        # if it isn't moved for a while we turn off mouse look.
        when_last_moved = time.time() - self.last_mouse_event_time

        if when_last_moved > 3.0:
            const.DISABLE_MOUSE_LOOK = 1

        # if we move the mouse enough mouse look will go back on.
        if self.last_mouse_moves > 20:
            self.last_mouse_moves = 0
            const.DISABLE_MOUSE_LOOK = 0

        if const.DISABLE_MOUSE_LOOK:
            tv.auto_scroll = True

        # print "autoscrol is :%s:" % tv.auto_scroll
        if tv.auto_scroll:
            if p.left - border < v.left:
                v.left = max(v.left - inc, p.left - border)
            if p.right + border > v.right:
                v.right = min(v.right + inc, p.right + border)
            if p.top - border < v.top:
                v.top = max(v.top - inc, p.top - border)
            if p.bottom + border > v.bottom:
                v.bottom = min(v.bottom + inc, p.bottom + border)

        b = tv.bounds
        # print "*"*30
        # print "b is:"
        # print b
        # print "v is:"
        # print v
        if b != None:
            v.left = max(v.left, b.left)
            v.right = min(v.right, b.right)
            v.top = max(v.top, b.top)
            v.bottom = min(v.bottom, b.bottom)

        updates = []

        s = screen.subsurface(S_VIEW)
        # s.fill((0,0,0))
        self.tv.paint(s)

        updates.append(S_VIEW)

        vv = len(self.tv.robots) - (self.tv.robots_total * (100 - self.percent) / 100)
        if vv != self.interface.stats_draw.robots_total:
            self.interface.stats_draw.robots_image = self.tv.images["robot"][0]
            self.interface.stats_draw.robots_total = vv
            # self.interface.dirty['background_illustration'] = 1
            self.interface.stats_draw.robots_update = 1

        #
        interface_rects = self.interface.update(tv, screen)
        self.interface.stats_draw.robots_update = 0
        updates.extend(interface_rects)

        if self.hair_update != None:
            updates.append(self.hair_update)
            self.hair_update = None

        if show_hairs == True:
            bmx, bmy = pygame.mouse.get_pos()

            if self.last_show_hairs != show_hairs:
                pass
                # pygame.mouse.set_visible(False)
                # pygame.mouse.set_pos((bmx,bmy))

            mx, my = pygame.mouse.get_pos()
            hairs = self.hairs
            hw, hh = hairs.get_width(), hairs.get_height()
            hr = pygame.Rect((mx - hw / 2, my - hh / 2, hw, hh))
            tmp = pygame.Surface((hw, hh), 0, 32)
            tmp.blit(screen, (0, 0), hr)
            screen.blit(hairs, hr)
            updates.append(hr)
        else:
            # bmx,bmy = pygame.mouse.get_pos()

            if self.last_show_hairs != show_hairs:
                pass
                # pygame.mouse.set_visible(True)
                # pygame.mouse.set_pos((bmx,bmy))

        pygame.display.update(updates)

        if show_hairs == True:
            screen.blit(tmp, hr)
            self.hair_update = hr
        self.last_show_hairs = show_hairs

    def update(self, screen):
        self.paint(screen)
        # [pygame.Rect(0,0,screen.get_width(),screen.get_height())]

    def loop(self):
        tv = self.tv

        # HACK for optimization of gfx
        tv.tiles[0].image = pygame.Surface((32, 32))

        self.tv.loop()

        tv.tiles[0].image = None
        pass

        if self.frames % (FPS * 12) == 0:
            self.interface.new_random_message()

        # if not self.frames % 5:
        #    flock.flock(tv.robots, 1, 1)
        # else:
        #    flock.flock(tv.robots, 0, 0)
        #    pass

        tv.robots.loop()  # flock 'em

        #         for r in tv.robots:
        #             r._frect = r.frect
        #             r.frect = pygame.Rect(r.rect)

        self.frames += 1

        # check to see if the beat the level...
        if (tv.robots_total - len(tv.robots)) * 100 / tv.robots_total >= self.percent:
            tv.castle.unit.stats["Health"] = 10.0
            tv.castle.backup_castle()
            self.game.robots_left = len(tv.robots)
            self.game.max_robots = self.interface.max_robots

            # self.interface.max_robots = len(tv.robots)

            from . import states

            self.game.data["levels"].append(self.round)
            if len(self.game.data["levels"]) == 8:
                return states.GameWon(self.game)

            return states.NextLevel(self.game)

    def event(self, e):
        #         if e.type == MOUSEMOTION and 1 in e.buttons:
        #             self.tv.view.x -= e.rel[0]
        #             self.tv.view.y -= e.rel[1]

        # if the castle is in a pickup state, then see if enter or right click.

        # note the time when the mouse is moved.
        if e.type in [MOUSEMOTION, MOUSEBUTTONDOWN, MOUSEBUTTONUP]:
            self.last_mouse_event_time = time.time()
            self.last_mouse_moves += 1

        self.interface.event(self.tv, e)

        self.tv.castle.event(self.tv, e)

        if e.type == KEYDOWN and e.key in (K_RETURN, K_p, K_h):
            from . import states

            return states.Pause(self.game, "pause", self)

        if e.type == KEYDOWN and e.key == K_ESCAPE:
            from . import states
            from . import menu

            return states.Prompt(self.game, "give up? y/n", menu.Menu(self.game), self)

        # CHEAT keys
        if e.type == KEYDOWN and e.key == K_F10:
            from . import menu
            from . import states

            return states.GameOver(self.game)
        if e.type == KEYDOWN and e.key in [K_F11, K_n]:
            from . import menu
            from . import states

            self.game.data["levels"].append(self.round)
            return states.NextLevel(self.game)
        if e.type == KEYDOWN and e.key in [K_F12, K_v]:
            from . import menu
            from . import states

            self.game.data["levels"].append(self.round)
            return states.GameWon(self.game)

        if e.type == KEYDOWN and e.key in [K_m]:
            const.DISABLE_MOUSE_LOOK = not const.DISABLE_MOUSE_LOOK

        # bound the mouse position


#         x,y = pygame.mouse.get_pos()
#         r = pygame.Rect(x,y,1,1)
#         r.clamp_ip(S_VIEW)
#         pygame.mouse.set_pos((r.x,r.y))



import pygame
import math


def scale_rect(r, orig_size, size):
    width, height = size
    owidth, oheight = orig_size
    sw = width / float(owidth)
    sh = height / float(oheight)
    return pygame.Rect(
        list(
            map(int, list(map(math.ceil, [r[0] * sw, r[1] * sh, r[2] * sw, r[3] * sh])))
        )
    )



# file: cycling list.
# purpose: a pointer to the current item is needed.
#     A goto_next method is needed.  Which cycles over the next idx.


try:
    from UserList import UserList
except ImportError:
    try:
        from collections import UserList
    except ImportError:
        UserList = list


class cyclic_list(UserList):
    def __init__(self, *args, **kwargs):
        UserList.__init__(self, *args, **kwargs)
        self.idx = 0
        if not hasattr(self, "data"):
            self.data = self

    def __iter__(self):
        self.idx = 0
        return self

    def __next__(self):
        if self.idx >= len(self.data):
            raise StopIteration
        result = self.data[self.idx]
        self.idx += 1
        return result

    def next(self):
        """increments the cursor, and returns the new current idx."""

        self.idx += 1

        if self.idx >= len(self.data):
            self.idx = 0

        return self.idx

    def nextone(self):
        i = self.next()
        return self.data[i]

    def prev(self):
        """decrements the cursor, and returns the new current idx."""
        self.idx -= 1

        if self.idx < 0:
            self.idx = len(self.data) - 1

        return self.idx

    def cur(self):
        """returns the element at the current cursor."""
        return self.data[self.idx]

    def set_cur(self, idx):
        """Sets the current cursor to the given idx."""
        self.idx = idx


if __name__ == "__main__":
    c = cyclic_list([1, 2, 3, 4])
    print((c.idx))
    next(c)
    print((c.idx))
    next(c)
    print((c.idx))
    next(c)
    print((c.idx))
    next(c)
    print((c.idx))

    for x in range(5):
        c.prev()
        print((c.idx))

    for x in range(5):
        print(c.nextone())



"""
"""
import os, re
import pygame

from .const import *
from . import surface


def _list_themes(dir):
    d = {}
    for entry in os.listdir(dir):
        if os.path.exists(os.path.join(dir, entry, "config.txt")):
            d[entry] = os.path.join(dir, entry)
    return d


class Theme:
    """Theme interface.

    <p>If you wish to create your own theme, create a class with this interface, and
    pass it to gui.App via <tt>gui.App(theme=MyTheme())</tt>.</p>

    <strong>Default Theme</strong>

    <pre>Theme(dirs='default')</pre>
    <dl>
    <dt>dirs<dd>Name of the theme dir to load a theme from.  May be an absolute path to a theme, if pgu is not installed, or if you created your own theme.  May include several dirs in a list if data is spread across several themes.
    </dl>

    <strong>Example</strong>

    <code>
    theme = gui.Theme("default")
    theme = gui.Theme(["mytheme","mytheme2"])
    </code>
    """

    def __init__(self, dirs="default"):
        self.config = {}
        self.dict = {}
        self._loaded = []
        self.cache = {}
        self._preload(dirs)
        pygame.font.init()

    def _preload(self, ds):
        if not isinstance(ds, list):
            ds = [ds]
        for d in ds:
            if d not in self._loaded:
                self._load(d)
            self._loaded.append(d)

    def _load(self, name):
        # theme_dir = themes[name]

        # try to load the local dir, or absolute path
        dnames = [name]

        dnames.append(os.path.join("data", "themes", name))
        # if the package isn't installed and people are just
        # trying out the scripts or examples
        dnames.append(
            os.path.join(os.path.dirname(__file__), "..", "..", "data", "themes", name)
        )

        # if the package is installed, and the package is installed
        # in /usr/lib/python2.3/site-packages/pgu/
        # or c:\python23\lib\site-packages\pgu\
        # the data is in ... lib/../share/ ...
        dnames.append(
            os.path.join(
                os.path.dirname(__file__),
                "..",
                "..",
                "..",
                "..",
                "share",
                "pgu",
                "themes",
                name,
            )
        )
        dnames.append(
            os.path.join(
                os.path.dirname(__file__),
                "..",
                "..",
                "..",
                "..",
                "..",
                "share",
                "pgu",
                "themes",
                name,
            )
        )

        for dname in dnames:
            if os.path.isdir(dname):
                break
        if not os.path.isdir(dname):
            raise "could not find theme " + name

        fname = os.path.join(dname, "config.txt")
        try:
            f = open(fname)
            for line in f.readlines():
                vals = line.strip().split()
                if len(vals) < 3:
                    continue
                cls = vals[0]
                del vals[0]
                pcls = ""
                if cls.find(":") >= 0:
                    cls, pcls = cls.split(":")
                attr = vals[0]
                del vals[0]
                self.config[cls + ":" + pcls + " " + attr] = (dname, vals)
        finally:
            f.close()

    is_image = re.compile("\.(gif|jpg|bmp|png|tga)$", re.I)

    def _get(self, key):
        if not key in self.config:
            return
        if key in self.dict:
            return self.dict[key]
        dvals = self.config[key]
        dname, vals = dvals
        # theme_dir = themes[name]
        v0 = vals[0]
        if v0[0] == "#":
            v = pygame.color.Color(v0)
        elif v0.endswith(".ttf") or v0.endswith(".TTF"):
            v = pygame.font.Font(os.path.join(dname, v0), int(vals[1]))
        elif self.is_image.search(v0) is not None:
            v = pygame.image.load(os.path.join(dname, v0))
        else:
            try:
                v = int(v0)
            except:
                v = pygame.font.SysFont(v0, int(vals[1]))
        self.dict[key] = v
        return v

    def get(self, cls, pcls, attr):
        """Interface method -- get the value of a style attribute.

        <pre>Theme.get(cls,pcls,attr): return value</pre>

        <dl>
        <dt>cls<dd>class, for example "checkbox", "button", etc.
        <dt>pcls<dd>pseudo class, for example "hover", "down", etc.
        <dt>attr<dd>attribute, for example "image", "background", "font", "color", etc.
        </dl>

        <p>returns the value of the attribute.</p>

        <p>This method is called from [[gui-style]].</p>
        """

        if not self._loaded:
            self._preload("default")

        o = cls + ":" + pcls + " " + attr

        # if not hasattr(self,'_count'):
        #    self._count = {}
        # if o not in self._count: self._count[o] = 0
        # self._count[o] += 1

        if o in self.cache:
            return self.cache[o]

        v = self._get(cls + ":" + pcls + " " + attr)
        if v:
            self.cache[o] = v
            return v

        pcls = ""
        v = self._get(cls + ":" + pcls + " " + attr)
        if v:
            self.cache[o] = v
            return v

        cls = "default"
        v = self._get(cls + ":" + pcls + " " + attr)
        if v:
            self.cache[o] = v
            return v

        v = 0
        self.cache[o] = v
        return v

    def box(self, w, s):
        style = w.style
        style.cache(1)

        c = (0, 0, 0)
        if style.border_color != 0:
            c = style.border_color
        w, h = s.get_width(), s.get_height()

        s.fill(c, (0, 0, w, style.border_top))
        s.fill(c, (0, h - style.border_bottom, w, style.border_bottom))
        s.fill(c, (0, 0, style.border_left, h))
        s.fill(c, (w - style.border_right, 0, style.border_right, h))

        style.cache(0)

    def getspacing(self, w):
        # return the top, right, bottom, left spacing around the widget
        if not hasattr(w, "_spacing"):  # HACK: assume spacing doesn't change re pcls
            s = w.style
            s.cache(1)
            xt = s.margin_top + s.border_top + s.padding_top
            xr = s.padding_right + s.border_right + s.margin_right
            xb = s.padding_bottom + s.border_bottom + s.margin_bottom
            xl = s.margin_left + s.border_left + s.padding_left
            s.cache(0)
            w._spacing = xt, xr, xb, xl
        return w._spacing

    def resize(self, w, m):
        def func(width=None, height=None):
            # ww,hh = m(width,height)
            ow, oh = width, height

            s = w.style

            s.cache(1)
            pt, pr, pb, pl = (
                s.padding_top,
                s.padding_right,
                s.padding_bottom,
                s.padding_left,
            )
            bt, br, bb, bl = (
                s.border_top,
                s.border_right,
                s.border_bottom,
                s.border_left,
            )
            mt, mr, mb, ml = (
                s.margin_top,
                s.margin_right,
                s.margin_bottom,
                s.margin_left,
            )
            s.cache(0)

            xt = pt + bt + mt
            xr = pr + br + mr
            xb = pb + bb + mb
            xl = pl + bl + ml
            ttw = xl + xr
            tth = xt + xb

            ww, hh = None, None
            if width != None:
                ww = width - ttw
            if height != None:
                hh = height - tth
            ww, hh = m(ww, hh)

            rect = pygame.Rect(0 + xl, 0 + xt, ww, hh)
            w._rect_content = rect  # pygame.Rect(0 + xl, 0 + xt, width, height)
            # r = rect

            if width == None:
                width = ww
            if height == None:
                height = hh
            # if the widget hasn't respected the style.width,
            # style height, we'll add in the space for it...
            width = max(width - ttw, ww, w.style.width)
            height = max(height - tth, hh, w.style.height)

            # width = max(ww,w.style.width-tw)
            # height = max(hh,w.style.height-th)

            r = pygame.Rect(rect.x, rect.y, width, height)

            w._rect_padding = pygame.Rect(
                r.x - pl, r.y - pt, r.w + pl + pr, r.h + pt + pb
            )
            r = w._rect_padding
            w._rect_border = pygame.Rect(
                r.x - bl, r.y - bt, r.w + bl + br, r.h + bt + bb
            )
            r = w._rect_border
            w._rect_margin = pygame.Rect(
                r.x - ml, r.y - mt, r.w + ml + mr, r.h + mt + mb
            )

            # align it within it's zone of power.
            dx = width - rect.w
            dy = height - rect.h
            # rect.x += (1)*dx/2
            # rect.y += (1)*dy/2
            rect.x += (w.style.align + 1) * dx / 2
            rect.y += (w.style.valign + 1) * dy / 2

            # print w,ow, w._rect_margin.w,  ttw
            return w._rect_margin.w, w._rect_margin.h

        return func

    def paint(self, w, m):
        def func(s):
            if w.disabled:
                orig = s
                s = pygame.Surface((s.get_width(), s.get_height()), 0, s)
                s.blit(orig, (0, 0))

            if hasattr(w, "background"):
                w.background.paint(surface.subsurface(s, w._rect_border))
            self.box(w, surface.subsurface(s, w._rect_border))
            r = m(surface.subsurface(s, w._rect_content))

            if w.disabled:
                s.set_alpha(128)
                orig.blit(s, (0, 0))

            return r

        return func

    def event(self, w, m):
        def func(e):
            rect = w._rect_content
            if e.type == MOUSEBUTTONUP or e.type == MOUSEBUTTONDOWN:
                sub = pygame.event.Event(
                    e.type,
                    {"button": e.button, "pos": (e.pos[0] - rect.x, e.pos[1] - rect.y)},
                )
            elif e.type == CLICK:
                sub = pygame.event.Event(
                    e.type,
                    {"button": e.button, "pos": (e.pos[0] - rect.x, e.pos[1] - rect.y)},
                )
            elif e.type == MOUSEMOTION:
                sub = pygame.event.Event(
                    e.type,
                    {
                        "buttons": e.buttons,
                        "pos": (e.pos[0] - rect.x, e.pos[1] - rect.y),
                        "rel": e.rel,
                    },
                )
            else:
                sub = e
            r = m(sub)
            return r

        return func

    def update(self, w, m):
        def func(s):
            if w.disabled:
                return []
            r = m(surface.subsurface(s, w._rect_content))
            if type(r) == list:
                dx, dy = w._rect_content.topleft
                for rr in r:
                    rr.x, rr.y = rr.x + dx, rr.y + dy
            return r

        return func

    def open(self, w, m):
        def func(widget=None, x=None, y=None):
            if not hasattr(w, "_rect_content"):
                (
                    w.rect.w,
                    w.rect.h,
                ) = w.resize()  # HACK: so that container.open won't resize again!
            rect = w._rect_content
            ##print w.__class__.__name__, rect
            if x != None:
                x += rect.x
            if y != None:
                y += rect.y
            return m(widget, x, y)

        return func

    # def open(self,w,m):
    #    def func(widget=None):
    #        return m(widget)
    #    return func

    def decorate(self, widget, level):
        """Interface method -- decorate a widget.

        <p>The theme system is given the opportunity to decorate a widget methods at the
        end of the Widget initializer.</p>

        <pre>Theme.decorate(widget,level)</pre>

        <dl>
        <dt>widget<dd>the widget to be decorated
        <dt>level<dd>the amount of decoration to do, False for none, True for normal amount, 'app' for special treatment of App objects.
        </dl>
        """

        w = widget
        if level == False:
            return

        if type(w.style.background) != int:
            w.background = Background(w, self)

        if level == "app":
            return

        for k, v in list(w.style.__dict__.items()):
            if k in ("border", "margin", "padding"):
                for kk in ("top", "bottom", "left", "right"):
                    setattr(w.style, "%s_%s" % (k, kk), v)

        w.paint = self.paint(w, w.paint)
        w.event = self.event(w, w.event)
        w.update = self.update(w, w.update)
        w.resize = self.resize(w, w.resize)
        w.open = self.open(w, w.open)

    def render(self, s, box, r):
        """Interface method - render a special widget feature.

        <pre>Theme.render(s,box,r)</pre>

        <dl>
        <dt>s<dt>pygame.Surface
        <dt>box<dt>box data, a value returned from Theme.get, typically a pygame.Surface
        <dt>r<dt>pygame.Rect with the size that the box data should be rendered
        </dl>

        """

        if box == 0:
            return

        if type(box) == tuple:
            s.fill(box, r)
            return

        x, y, w, h = r.x, r.y, r.w, r.h
        ww, hh = box.get_width() / 3, box.get_height() / 3
        xx, yy = x + w, y + h
        src = pygame.rect.Rect(0, 0, ww, hh)
        dest = pygame.rect.Rect(0, 0, ww, hh)

        s.set_clip(pygame.Rect(x + ww, y + hh, w - ww * 2, h - hh * 2))
        src.x, src.y = ww, hh
        for dest.y in range(y + hh, yy - hh, hh):
            for dest.x in range(x + ww, xx - ww, ww):
                s.blit(box, dest, src)

        s.set_clip(pygame.Rect(x + ww, y, w - ww * 3, hh))
        src.x, src.y, dest.y = ww, 0, y
        for dest.x in range(x + ww, xx - ww * 2, ww):
            s.blit(box, dest, src)
        dest.x = xx - ww * 2
        s.set_clip(pygame.Rect(x + ww, y, w - ww * 2, hh))
        s.blit(box, dest, src)

        s.set_clip(pygame.Rect(x + ww, yy - hh, w - ww * 3, hh))
        src.x, src.y, dest.y = ww, hh * 2, y + h - hh
        for dest.x in range(x + ww, xx - ww * 2, ww):
            s.blit(box, dest, src)
        dest.x = xx - ww * 2
        s.set_clip(pygame.Rect(x + ww, yy - hh, w - ww * 2, hh))
        s.blit(box, dest, src)

        s.set_clip(pygame.Rect(x, y + hh, xx, h - hh * 3))
        src.y, src.x, dest.x = hh, 0, x
        for dest.y in range(y + hh, yy - hh * 2, hh):
            s.blit(box, dest, src)
        dest.y = yy - hh * 2
        s.set_clip(pygame.Rect(x, y + hh, xx, h - hh * 2))
        s.blit(box, dest, src)

        s.set_clip(pygame.Rect(xx - ww, y + hh, xx, h - hh * 3))
        src.y, src.x, dest.x = hh, ww * 2, x + w - ww
        for dest.y in range(y + hh, yy - hh * 2, hh):
            s.blit(box, dest, src)
        dest.y = yy - hh * 2
        s.set_clip(pygame.Rect(xx - ww, y + hh, xx, h - hh * 2))
        s.blit(box, dest, src)

        s.set_clip()
        src.x, src.y, dest.x, dest.y = 0, 0, x, y
        s.blit(box, dest, src)

        src.x, src.y, dest.x, dest.y = ww * 2, 0, x + w - ww, y
        s.blit(box, dest, src)

        src.x, src.y, dest.x, dest.y = 0, hh * 2, x, y + h - hh
        s.blit(box, dest, src)

        src.x, src.y, dest.x, dest.y = ww * 2, hh * 2, x + w - ww, y + h - hh
        s.blit(box, dest, src)


import pygame
from . import widget


class Background(widget.Widget):
    def __init__(self, value, theme, **params):
        params["decorate"] = False
        widget.Widget.__init__(self, **params)
        self.value = value
        self.theme = theme

    def paint(self, s):
        r = pygame.Rect(0, 0, s.get_width(), s.get_height())
        v = self.value.style.background
        if type(v) == tuple:
            s.fill(v)
        else:
            self.theme.render(s, v, r)



import pygame
from . import sound_info

PRINTDEBUG = 0


def tile_block(g, t, s):
    a = s

    c = t.config
    if c["top"] == 1 and a._rect.bottom <= t._rect.top and a.rect.bottom > t.rect.top:
        a.rect.bottom = t.rect.top
    if c["left"] == 1 and a._rect.right <= t._rect.left and a.rect.right > t.rect.left:
        a.rect.right = t.rect.left
    if c["right"] == 1 and a._rect.left >= t._rect.right and a.rect.left < t.rect.right:
        a.rect.left = t.rect.right
    if (
        c["bottom"] == 1
        and a._rect.top >= t._rect.bottom
        and a.rect.top < t.rect.bottom
    ):
        a.rect.top = t.rect.bottom


class WallBlast:
    def __init__(self):
        from . import explode, steam

        # self.e1 = explode.Effect(64,16)
        # self.e2 = steam.Effect(128,32,24,0)
        self.e2 = steam.Effect(120, 40, 16, 0, (143, 136, 129))

    def loop(self, pos):
        # self.e1.loop(pos)
        self.e2.loop(pos)

    def paint(self, screen, origin):
        # self.e1.paint(screen,origin)
        self.e2.paint(screen, origin)


def tile_wall(g, t, s):
    # when hit, this cracks

    if s not in g.sprites:
        return
    g.sprites.remove(s)

    g.level.game.sm.Play(sound_info.hitwall.nextone())

    tx, ty = t.tx, t.ty
    for a, dx, dy in [(2, 0, 0), (1, -1, 0), (1, 1, 0), (1, 0, 1), (1, 0, -1)]:
        v = g.get((tx + dx, ty + dy))
        if v in (4, 5, 6):
            v = min(7, v + a)
            g.set((tx + dx, ty + dy), v)

    from . import effect

    tw, th = g.iso_w, g.iso_h
    e = effect.effect_new(
        g, pygame.Rect(t.tx * tw + tw / 2, t.ty * th + th / 2, 1, 1), WallBlast(), 20
    )
    e.z = s.z + 16


def tile_coal(g, t, s):
    # print 'you hit coal'
    pass

    g.level.game.sm.Play(sound_info.coal.nextone())

    s.check_for_pickup(g, t)


def tile_water(g, t, s):
    # print 'you hit water'
    pass
    g.level.game.sm.Play(sound_info.water.nextone())

    s.check_for_pickup(g, t)


def tile_rubble(g, t, s):
    print("you hit rubble")

    s.check_for_pickup(g, t)


def tile_part(g, t, s):
    # print "*"*30
    # print type(s)
    # print s
    # print g
    if PRINTDEBUG:
        print("generic tile part")
    s.check_for_pickup(g, t)


def tile_cannon(g, t, s):
    if PRINTDEBUG:
        print("you hit cannon part")
    s.check_for_pickup(g, t)


def tile_engine(g, t, s):
    if PRINTDEBUG:
        print("you hit engine part")
    s.check_for_pickup(g, t)


def tile_watertank(g, t, s):
    if PRINTDEBUG:
        print("you hit watertank part")
    s.check_for_pickup(g, t)


def tile_coalstorageroom(g, t, s):
    if PRINTDEBUG:
        print("you hit coalstorageroom part")
    s.check_for_pickup(g, t)


def tile_armour(g, t, s):
    if PRINTDEBUG:
        print("you hit armour part")
    s.check_for_pickup(g, t)


def tile_limit(g, t, s):
    print((s, "out of bounds"))
    if s in g.sprites:
        g.sprites.remove(s)
    # HACK
    if s in g.robots:
        g.robots.remove(s)
    s.state = "dead"



"""a html renderer
"""
import htmllib
import re
import pygame
from pygame.locals import *

from .pgu import gui

_amap = {
    "left": -1,
    "right": 1,
    "center": 0,
    None: None,
    "": None,
}
_vamap = {
    "top": -1,
    "bottom": 1,
    "center": 0,
    "middle": 0,
    None: None,
    "": None,
}


class _dummy:
    pass


class _flush:
    def __init__(self):
        self.style = _dummy()
        self.style.font = None
        self.style.color = None
        self.cls = None

    def add(self, w):
        pass

    def space(self, v):
        pass


class _hr(gui.Color):
    def __init__(self, **params):
        gui.Color.__init__(self, (0, 0, 0), **params)

    def resize(self, width=None, height=None):
        w, h = self.style.width, self.style.height
        # if width != None: self.rect.w = width
        # else: self.rect.w = 1

        # xt,xr,xb,xl = self.getspacing()

        if width != None:
            w = max(w, width)
        if height != None:
            h = max(h, height)
        w = max(w, 1)
        h = max(h, 1)

        return w, h  # self.container.rect.w,h

        # self.rect.w = max(1,width,self.container.rect.w-(xl+xr))

        # print self.rect
        # self.rect.w = 1


class _html(htmllib.HTMLParser):
    def init(self, doc, font, color, _globals, _locals):
        self.mystack = []
        self.document = doc
        self.myfont = self.font = font
        self.mycolor = self.color = color
        self.myopen("document", self.document)

        self.form = None

        self._globals = _globals
        self._locals = _locals

    def myopen(self, type_, w):
        self.mystack.append((type_, w))
        self.type, self.item = type_, w

        # self.font = self.item.style.font
        # self.color = self.item.style.color

        if not self.font:
            self.font = self.myfont
        if not self.color:
            self.color = self.mycolor

    def myclose(self, type_):
        t = None
        self.mydone()
        while t != type_:
            # if len(self.mystack)==0: return
            t, w = self.mystack.pop()
        t, w = self.mystack.pop()
        self.myopen(t, w)

    def myback(self, type_):
        if type(type_) == str:
            type_ = [
                type_,
            ]
        self.mydone()
        # print 'myback',type_
        t = None
        while t not in type_:
            # if len(self.mystack)==0: return
            t, w = self.mystack.pop()
        self.myopen(t, w)

    def mydone(self):
        # clearing out the last </p>
        if not hasattr(self.item, "layout"):
            return
        if len(self.item.layout._widgets) == 0:
            return
        w = self.item.layout._widgets[-1]
        if type(w) == tuple:
            del self.item.layout._widgets[-1]

    def start_b(self, attrs):
        self.font.set_bold(1)

    def end_b(self):
        self.font.set_bold(0)

    def start_i(self, attrs):
        self.font.set_italic(1)

    def end_i(self):
        self.font.set_italic(0)

    def start_u(self, attrs):
        self.font.set_underline(1)

    def end_u(self):
        self.font.set_underline(0)

    def start_br(self, attrs):
        self.do_br(attrs)

    def do_br(self, attrs):
        self.item.br(self.font.size(" ")[1])

    def attrs_to_map(self, attrs):
        k = None
        r = {}
        for k, v in attrs:
            r[k] = v
        return r

    def map_to_params(self, r):
        anum = re.compile("\D")

        params = {"style": {}}
        style = params["style"]

        if "bgcolor" in r:
            style["background"] = pygame.Color(r["bgcolor"])
        if "background" in r:
            style["background"] = pygame.image.load(r["background"])
        if "border" in r:
            style["border"] = int(r["border"])

        for k in ["width", "height", "colspan", "rowspan", "size", "min", "max"]:
            if k in r:
                params[k] = int(anum.sub("", r[k]))

        for k in ["name", "value"]:
            if k in r:
                params[k] = r[k]

        if "class" in r:
            params["cls"] = r["class"]

        if "align" in r:
            params["align"] = _amap[r["align"]]
        if "valign" in r:
            params["valign"] = _vamap[r["valign"]]

        if "style" in r:
            for st in r["style"].split(";"):
                # print st
                if ":" in st:
                    # print st.split(":")
                    k, v = st.split(":")
                    k = k.replace("-", "_")
                    k = k.replace(" ", "")
                    v = v.replace(" ", "")
                    if k == "color" or k == "border_color" or k == "background":
                        v = pygame.Color(v)
                    else:
                        v = int(anum.sub("", v))
                    style[k] = v
        return params

    def map_to_connects(self, e, r):
        for k, evt in [
            ("onclick", gui.CLICK),
            ("onchange", gui.CHANGE),
        ]:  # blah blah blah
            if k in r:
                # print k,r[k]
                e.connect(evt, self.myexec, (e, r[k]))

    def start_p(self, attrs):
        r = self.attrs_to_map(attrs)
        align = r.get("align", "left")

        self.check_p()
        self.item.block(_amap[align])

    def check_p(self):
        if len(self.item.layout._widgets) == 0:
            return
        if type(self.item.layout._widgets[-1]) == tuple:
            w, h = self.item.layout._widgets[-1]
            if w == 0:
                return
        self.do_br(None)

    def end_p(self):
        # print 'end p'
        self.check_p()

    def start_block(self, t, attrs, align=-1):
        r = self.attrs_to_map(attrs)
        params = self.map_to_params(r)
        if "cls" in params:
            params["cls"] = t + "." + params["cls"]
        else:
            params["cls"] = t
        b = gui.Document(**params)
        if "align" in params:
            align = params["align"]
        self.item.block(align)
        self.item.add(b)
        self.myopen(t, b)

    def end_block(self, t):
        self.myclose(t)
        self.item.block(-1)

    def start_div(self, attrs):
        self.start_block("div", attrs)

    def end_div(self):
        self.end_block("div")

    def start_center(self, attrs):
        self.start_block("div", attrs, 0)

    def end_center(self):
        self.end_block("div")

    def start_h1(self, attrs):
        self.start_block("h1", attrs)

    def end_h1(self):
        self.end_block("h1")

    def start_h2(self, attrs):
        self.start_block("h2", attrs)

    def end_h2(self):
        self.end_block("h2")

    def start_h3(self, attrs):
        self.start_block("h3", attrs)

    def end_h3(self):
        self.end_block("h3")

    def start_h4(self, attrs):
        self.start_block("h4", attrs)

    def end_h4(self):
        self.end_block("h4")

    def start_h5(self, attrs):
        self.start_block("h5", attrs)

    def end_h5(self):
        self.end_block("h5")

    def start_h6(self, attrs):
        self.start_block("h6", attrs)

    def end_h6(self):
        self.end_block("h6")

    def start_ul(self, attrs):
        self.start_block("ul", attrs)

    def end_ul(self):
        self.end_block("ul")

    def start_ol(self, attrs):
        self.start_block("ol", attrs)
        self.item.counter = 0

    def end_ol(self):
        self.end_block("ol")

    def start_li(self, attrs):
        self.myback(["ul", "ol"])
        cur = self.item
        self.start_block("li", attrs)
        if hasattr(cur, "counter"):
            cur.counter += 1
            self.handle_data("%d. " % cur.counter)
        else:
            self.handle_data("- ")

    # def end_li(self): self.end_block('li') #this isn't needed because of how the parser works

    def start_pre(self, attrs):
        self.start_block("pre", attrs)

    def end_pre(self):
        self.end_block("pre")

    def start_code(self, attrs):
        self.start_block("code", attrs)

    def end_code(self):
        self.end_block("code")

    def start_table(self, attrs):
        r = self.attrs_to_map(attrs)
        params = self.map_to_params(r)

        align = r.get("align", "left")
        self.item.block(_amap[align])

        t = gui.Table(**params)
        self.item.add(t)

        self.myopen("table", t)

    def start_tr(self, attrs):
        self.myback("table")
        self.item.tr()

    def _start_td(self, t, attrs):
        r = self.attrs_to_map(attrs)
        params = self.map_to_params(r)
        if "cls" in params:
            params["cls"] = t + "." + params["cls"]
        else:
            params["cls"] = t
        b = gui.Document(cls=t)

        self.myback("table")
        self.item.td(b, **params)
        self.myopen(t, b)

        # self.font = self.item.style.font
        # self.color = self.item.style.color

    def start_td(self, attrs):
        self._start_td("td", attrs)

    def start_th(self, attrs):
        self._start_td("th", attrs)

    def end_table(self):
        self.myclose("table")
        self.item.block(-1)

    def start_form(self, attrs):
        r = self.attrs_to_map(attrs)
        e = self.form = gui.Form()
        e.groups = {}

        self._locals[r.get("id", None)] = e

    def start_input(self, attrs):
        r = self.attrs_to_map(attrs)
        params = self.map_to_params(r)  # why bother
        # params = {}

        type_, name, value = (
            r.get("type", "text"),
            r.get("name", None),
            r.get("value", None),
        )
        f = self.form
        if type_ == "text":
            e = gui.Input(**params)
            self.map_to_connects(e, r)
            self.item.add(e)
        elif type_ == "radio":
            if name not in f.groups:
                f.groups[name] = gui.Group(name=name)
            g = f.groups[name]
            del params["name"]
            e = gui.Radio(group=g, **params)
            self.map_to_connects(e, r)
            self.item.add(e)
            if "checked" in r:
                g.value = value
        elif type_ == "checkbox":
            if name not in f.groups:
                f.groups[name] = gui.Group(name=name)
            g = f.groups[name]
            del params["name"]
            e = gui.Checkbox(group=g, **params)
            self.map_to_connects(e, r)
            self.item.add(e)
            if "checked" in r:
                g.value = value

        elif type_ == "button":
            e = gui.Button(**params)
            self.map_to_connects(e, r)
            self.item.add(e)
        elif type_ == "submit":
            e = gui.Button(**params)
            self.map_to_connects(e, r)
            self.item.add(e)
        elif type_ == "file":
            e = gui.Input(**params)
            self.map_to_connects(e, r)
            self.item.add(e)
            b = gui.Button(value="Browse...")
            self.item.add(b)

            def _browse(value):
                d = gui.FileDialog()
                d.connect(gui.CHANGE, gui.action_setvalue, (d, e))
                d.open()

            b.connect(gui.CLICK, _browse, None)

        self._locals[r.get("id", None)] = e

    def start_object(self, attrs):
        r = self.attrs_to_map(attrs)
        params = self.map_to_params(r)
        code = "e = %s(**params)" % r["type"]
        # print code
        # print params
        exec(code)
        # print e
        # print e.style.width,e.style.height
        self.map_to_connects(e, r)
        self.item.add(e)

        self._locals[r.get("id", None)] = e

    def start_select(self, attrs):
        r = self.attrs_to_map(attrs)
        params = {}

        name, value = r.get("name", None), r.get("value", None)
        e = gui.Select(name=name, value=value, **params)
        self.map_to_connects(e, r)
        self.item.add(e)
        self.myopen("select", e)

    def start_option(self, attrs):
        r = self.attrs_to_map(attrs)
        params = {}  # style = self.map_to_style(r)

        self.myback("select")
        e = gui.Document(**params)
        self.item.add(e, value=r.get("value", None))
        self.myopen("option", e)

    def end_select(self):
        self.myclose("select")

    def start_hr(self, attrs):
        self.do_hr(attrs)

    def do_hr(self, attrs):
        h = self.font.size(" ")[1] / 2

        r = self.attrs_to_map(attrs)
        params = self.map_to_params(r)
        params["style"]["padding"] = h
        print(params)

        self.item.block(0)
        self.item.add(_hr(**params))
        self.item.block(-1)

    def anchor_begin(self, href, name, type_):
        pass

    def anchor_end(self):
        pass

    def start_title(self, attrs):
        self.myopen("title", _flush())

    def end_title(self):
        self.myclose("title")

    def myexec(self, value):
        w, code = value
        g = self._globals
        l = self._locals
        l["self"] = w
        exec(code, g, l)

    def handle_image(self, src, alt, ismap, align, width, height):
        try:
            w = gui.Image(pygame.image.load(src))
            if align != "":
                self.item.add(w, _amap[align])
            else:
                self.item.add(w)
        except:
            print(("handle_image: missing %s" % src))

    def handle_data(self, txt):
        if self.type == "table":
            return
        elif self.type in ("pre", "code"):
            txt = txt.replace("\t", "        ")
            ss = txt.split("\n")
            if ss[-1] == "":
                del ss[-1]
            for sentence in ss:
                img = self.myfont.render(sentence, 1, self.mycolor)
                w = gui.Image(img)
                self.item.add(w)
                self.item.block(-1)
            return

        txt = re.compile("^[\t\r\n]+").sub("", txt)
        txt = re.compile("[\t\r\n]+$").sub("", txt)

        tst = re.compile("[\t\r\n]+").sub("", txt)
        if tst == "":
            return

        txt = re.compile("\s+").sub(" ", txt)
        if txt == "":
            return

        if txt == " ":
            self.item.space(self.myfont.size(" "))
            return

        for word in txt.split(" "):
            word = word.replace(chr(160), " ")  # &nbsp;
            # print self.item.cls
            w = gui.Image(self.myfont.render(word, 1, self.mycolor))
            self.item.add(w)
            self.item.space(self.myfont.size(" "))


class HTML(gui.Document):
    """a gui HTML object

    <pre>HTML(data,globals=None,locals=None)</pre>

    <dl>
    <dt>data <dd>html data
    <dt>globals <dd>global variables (for scripting)
    <dt>locals <dd>local variables (for scripting)
    </dl>

    <p>you may access html elements that have an id via widget[id]</p>
    """

    def __init__(self, data, globals=None, locals=None, **params):
        gui.Document.__init__(self, **params)

        _globals, _locals = globals, locals

        if _globals == None:
            _globals = {}
        if _locals == None:
            _locals = {}
        self._globals = _globals
        self._locals = _locals

        # font = gui.theme.get("label","","font")
        p = _html(htmllib.AS_IS, 0)
        print((self.style.color))
        p.init(self, self.style.font, self.style.color, _globals, _locals)
        p.feed(data)
        p.close()
        p.mydone()

    def __getitem__(self, k):
        return self._locals[k]


def render(font, rect, text, aa, color, bgcolor=(0, 0, 0, 0)):
    """render some html

    <pre>render(font,rect,text,aa,color,bgcolor=(0,0,0,0))</pre>
    """
    fnt, r, txt, a, fg, bg = font, rect, text, aa, color, bgcolor

    e = HTML(txt, font=fnt, color=fg)
    e.resize(width=rect.w)
    s = pygame.Surface((e.rect.w, e.rect.h), SWSURFACE | SRCALPHA, 32)
    s.fill(bg)
    e.paint(s)

    return s


def rendertrim(font, rect, text, aa, color, bgcolor=(0, 0, 0, 0)):
    """render html, and make sure to trim the size

    <pre>rendertrim(font,rect,text,aa,color,bgcolor=(0,0,0,0))</pre>
    """
    fnt, r, txt, a, fg, bg = font, rect, text, aa, color, bgcolor
    # print r
    w = HTML(txt, font=fnt, color=fg)
    w.resize(width=rect.w)
    s = pygame.Surface((w.rect.w, w.rect.h), SWSURFACE | SRCALPHA, 32)
    s.fill(bg)
    w.paint(s)

    minx, miny, maxx, maxy = 1024, 1024, -1024, -1024
    for e in w.layout.widgets:
        x, y, w, h = e.rect.x, e.rect.y, e.rect.w, e.rect.h
        minx = min(minx, x)
        miny = min(miny, y)
        x, y = x + w, y + h
        maxx = max(maxx, x)
        maxy = max(maxy, y)

    r = pygame.Rect(minx, miny, maxx - minx, maxy - miny)

    return s.subsurface((r))


def write(s, font, rect, text, aa=0, color=(0, 0, 0)):
    """write html to a surface

    <pre>write(s,font,rect,text,aa=0,color=(0,0,0))</pre>
    """
    fnt, r, txt, a, fg = font, rect, text, aa, color

    e = HTML(txt, font=fnt, color=fg)

    e.resize(width=rect.w)
    s = s.subsurface(rect)
    e.paint(s)


# vim: set filetype=python sts=4 sw=4 noet si :



"""

to keep the state of the castle.

"""
import random, pprint, copy

from . import effect, steam, explode


import pygame
from pygame.locals import *

from . import isovid
from . import robot


def dist(a, b):
    return abs(a[0] - b[0]) + abs(a[1] - b[1])


def sign(v):
    if v == 0:
        return v
    return v // abs(v)


from . import algo
from .const import *

# from units import *
from . import units
from . import items

from . import sound_info


def iso_facing_to_screen_facing(kx, ky):
    if kx == 0 and ky == -1:
        kx = -1
        ky = -1

    elif kx == 0 and ky == 1:
        kx = 1
        ky = 1

    elif kx == 1 and ky == -1:
        kx = 0
        ky = -1

    elif kx == -1 and ky == -1:
        kx = -1
        ky = 0
    elif kx == -1 and ky == 1:
        kx = 0
        ky = 1
    elif kx == 1 and ky == 1:
        kx = 1
        ky = 0

    elif kx == 1 and ky == 0:
        kx = 1
        ky = -1

    elif kx == -1 and ky == 0:
        kx = -1
        ky = 1
    return (kx, ky)


class CastleSprite(isovid.Sprite):
    def __init__(self, *args, **kwargs):
        isovid.Sprite.__init__(self, *args, **kwargs)

        self.unit = units.Castle()

        self.can_move = True
        self.loop = castle_loop
        self.hit = castle_hit

        self.doing = ""

        self.frame = 0

        # counters for when hitting coal.
        self.last_hit_coal = 0
        self.last_loop = 0

        # how long ago we updated steam.
        self.last_update_steam = 0

        self.path = []
        self.target = None

        self.fire_state = ""

        self.direction_dx = 0
        self.direction_dy = 0
        self.last_direction_dy = 0
        self.last_direction_dx = 0

    def no_move(self):
        self.can_move = False

    def yes_move(self):
        self.can_move = True

    def yes_pickup(self, e):
        print("yes pickup")

        self.doing = ""
        self.yes_move()

        if not hasattr(self.item_tile, "item"):
            print("WEIRD: tile does not have an item attached?")
            print((self.item_tile))
            print((dir(self.item_tile)))
            return

        print((self.item_tile.item))

        # if not pickup item, then it's not picked up.
        #  this can happen if coal/water tanks are too small.
        if self.unit.pickup_item(self.item_tile.item):
            item = self.item_tile.item

            if item.type & items.ITEM_PART:
                self.upgrade_something(item.name)

            if not AUTO_PICKUP:
                # remove the buttons.
                self.g.app.main_container.remove(self.g.app.question_table)

            # tw,th = self.g.iso_w,self.g.iso_h
            tx, ty = self.item_tile.tx, self.item_tile.ty
            self.g.set((tx, ty), 0)

    def no_pickup(self, e):
        if not AUTO_PICKUP:
            print("no pickup")
            # self.unit.pickup_item(self.item_sprite)
            self.doing = ""
            self.yes_move()
            # remove the buttons... if they are there.
            try:
                self.g.app.main_container.remove(self.g.app.question_table)
            except ValueError:
                pass

    def check_for_pickup(self, g, t):
        self.last_hit_coal += 1

        if AUTO_PICKUP:
            self.doing = "deciding on picking up"
            self.item_tile = t

            self.yes_pickup("")
        else:
            if not self.doing:
                print("adding button")
                # s.no_move()

                self.doing = "deciding on picking up"
                self.item_tile = t

                # TODO: show the question.

    def upgrade_something(self, upgrade_what):
        """this upgrades something."""

        amount = self.unit.stats["upgrade_amounts"][upgrade_what].cur()
        next(self.unit.stats["upgrade_amounts"][upgrade_what])
        self.unit.upgrade_part(upgrade_what, amount)

        # play upgrade sound here.
        self.g.level.game.sm.Play(sound_info.get_upgrade_sound(upgrade_what))

        # update the equipment section
        self.g.level.interface.new_equipment(self.unit.stats)

        # say a message about the upgrade.
        self.g.level.interface.new_upgrade_message(upgrade_what)

    def event(self, g, e):
        tw, th = g.iso_w, g.iso_h
        s = self
        sx, sy = s.rect.centerx // tw, s.rect.centery // th

        if hasattr(e, "pos"):
            epos = e.pos[0] - S_VIEW.x, e.pos[1] - S_VIEW.y
            # print epos
            # print "sx:%s  sy:%s" % (sx, sy)

        if e.type == MOUSEBUTTONDOWN and e.button == 3:
            tx, ty = g.screen_to_tile(epos)
            # print 'left',tx,ty
            w, h = g.size
            if (tx, ty) != (sx, sy) and (tx >= 0 and ty >= 0 and tx < w and ty < h):
                self.target = tx, ty
                self.path = algo.astar((sx, sy), (tx, ty), g.castle_layer, dist)
                # print self.path
            # elif self.doing:
            #    self.yes_pickup(e)

        # print e
        #        if (e.type == MOUSEBUTTONDOWN and e.button == 3 or
        #            e.type == JOYBUTTONDOWN and e.button ==JOY_FIRE_BUTTON):
        if e.type == MOUSEBUTTONDOWN and e.button == 1:
            # use up some steam in order to fire.
            self.unit.prep_fire()

        if e.type == MOUSEBUTTONUP and e.button == 1:
            tx, ty = g.screen_to_tile(epos)
            # print 'right',tx,ty

            x, y = g.screen_to_tile(epos)
            cannon_pressure = self.unit.try_do_fire()

            tw, th = g.iso_w, g.iso_h
            cball_new(
                g,
                self.rect,
                (x * tw + tw // 2, y * th + th // 2),
                self.unit.stats["Damage"],
                cannon_pressure * 1.5,
            )

        if e.type == JOYBUTTONDOWN:
            pass

        # if self.direction_dy != 0:
        #    self.last_direction_dy = self.direction_dy
        # if self.direction_dx != 0:
        #    self.last_direction_dx = self.direction_dx

        if e.type == JOYAXISMOTION:
            if e.axis == 0:
                if round(e.value) < 0:
                    self.direction_dx = -1
                    self.last_direction_dx = self.direction_dx
                    self.last_direction_dy = self.direction_dy
                    # print "less than"
                elif round(e.value) > 0:
                    self.direction_dx = 1
                    self.last_direction_dx = self.direction_dx
                    self.last_direction_dy = self.direction_dy
                    # print "greater than"
                else:
                    # print "nothing!"
                    self.last_direction_dx = self.direction_dx
                    self.direction_dx = 0

            if e.axis == 1:
                if round(e.value) < 0:
                    self.direction_dy = -1
                    self.last_direction_dy = self.direction_dy
                    self.last_direction_dx = self.direction_dx
                elif round(e.value) > 0:
                    self.direction_dy = 1
                    self.last_direction_dy = self.direction_dy
                    self.last_direction_dx = self.direction_dx
                else:
                    self.last_direction_dy = self.direction_dy
                    self.direction_dy = 0

        if e.type == JOYBUTTONUP and e.button == JOY_FIRE_BUTTON:
            pass

            if 0:
                self.direction_dx = 0
                self.direction_dy = 0
                tx, ty = g.screen_to_tile(epos)
                x, y = g.screen_to_tile(epos)

                if self.unit.try_fire():
                    # print (x*tw+tw/2,y*th+th/2)
                    cball_new(
                        g,
                        s.rect,
                        (x * tw + tw // 2, y * th + th // 2),
                        self.unit.stats["Damage"],
                    )

        if e.type == JOYBUTTONDOWN and e.button == JOY_FIRE_BUTTON:
            # TODO: find out which way we are facing.
            # TODO: fire in facing direction.
            self.fire_state = "firing"

        if (e.type == JOYBUTTONDOWN and e.button == JOY_FIRE_BUTTON) or (
            e.type == KEYDOWN and e.key in [K_f, K_LCTRL, K_RCTRL, K_SPACE]
        ):
            sx, sy = self.rect.centerx // tw, self.rect.centery // th

            # print "last dx:%s:   last dy:%s:" % (self.last_direction_dx, self.last_direction_dy)
            # print "sx:%s:   sy:%s:" % (sx, sy)

            self.unit.prep_fire()

        if (e.type == JOYBUTTONUP and e.button == JOY_FIRE_BUTTON) or (
            e.type == KEYUP and e.key in [K_f, K_LCTRL, K_RCTRL, K_SPACE]
        ):
            facing_x, facing_y = (self.last_direction_dx, self.last_direction_dy)
            facing_x, facing_y = iso_facing_to_screen_facing(
                self.last_direction_dx, self.last_direction_dy
            )

            x = (5 * facing_x) + sx
            y = (5 * facing_y) + sy

            # x,y = g.screen_to_tile(epos)
            cannon_pressure = self.unit.try_do_fire()

            tw, th = g.iso_w, g.iso_h
            cball_new(
                g,
                self.rect,
                (x * tw + tw // 2, y * th + th // 2),
                self.unit.stats["Damage"],
                cannon_pressure,
            )

            # use up some steam in order to fire.
            # if self.unit.try_fire():
            #    cball_new(g,s.rect,(x*tw+tw/2,y*th+th/2), self.unit.stats['Damage'])

        if self.doing:
            if (
                (e.type == KEYDOWN and self.doing and e.key == K_RETURN)
                or (e.type == KEYDOWN and self.doing and e.key == K_SPACE)
                or (e.type == MOUSEBUTTONDOWN and e.button == 3)
                or (e.type == JOYBUTTONDOWN and e.button == JOY_PICKUP_BUTTON)
            ):
                self.yes_pickup(e)

        if e.type == KEYUP:
            if e.key in [K_UP, K_w]:
                self.last_direction_dy = self.direction_dy
                self.direction_dy = 0
            elif e.key in [K_DOWN, K_s]:
                self.last_direction_dy = self.direction_dy
                self.direction_dy = 0
            elif e.key in [K_LEFT, K_a]:
                self.last_direction_dx = self.direction_dx
                self.direction_dx = 0
            elif e.key in [K_RIGHT, K_d]:
                self.last_direction_dx = self.direction_dx
                self.direction_dx = 0

        # some testing keys.
        if e.type == KEYDOWN:
            if e.key in [K_UP, K_w]:
                self.direction_dy = -1
                self.last_direction_dy = self.direction_dy
                self.last_direction_dx = self.direction_dx
            elif e.key in [K_DOWN, K_s]:
                self.direction_dy = 1
                self.last_direction_dy = self.direction_dy
                self.last_direction_dx = self.direction_dx
            elif e.key in [K_LEFT, K_a]:
                self.direction_dx = -1
                self.last_direction_dx = self.direction_dx
                self.last_direction_dy = self.direction_dy
            elif e.key in [K_RIGHT, K_d]:
                self.direction_dx = 1
                self.last_direction_dx = self.direction_dx
                self.last_direction_dy = self.direction_dy

            if e.key == K_KP_PLUS:
                self.unit.stats["Speed"] += 1

            elif e.key == K_KP_MINUS:
                self.unit.stats["Speed"] -= 1

            elif e.key == K_h:
                self.unit.hit(2)

            elif e.key == K_r:
                # tx, ty = self.item_tile.tx, self.item_tile.ty
                # drop rubble test.
                tx, ty = self.rect.centerx // tw, self.rect.centery // th

                self.g.set((tx, ty), 7)

            if UPGRADE_FUN:
                if e.key == K_1:
                    self.upgrade_something("UpEngine Efficiency")
                elif e.key == K_2:
                    self.upgrade_something("UpEngine Speed")
                elif e.key == K_3:
                    self.upgrade_something("UpCannon Balls")
                elif e.key == K_4:
                    self.upgrade_something("UpCannon Power")
                elif e.key == K_5:
                    self.upgrade_something("UpArmour")
                elif e.key == K_6:
                    self.upgrade_something("UpSteam Tank")
                elif e.key == K_7:
                    self.upgrade_something("UpWater Tank")
                elif e.key == K_8:
                    self.upgrade_something("UpCoal Tank")

                elif e.key in [K_9, K_0]:
                    pprint.pprint(self.unit.stats)

            if e.key == K_y:
                # get_savable_upgrade_amounts()
                print((self.backup_castle()))
            if e.key == K_u:
                # get_savable_upgrade_amounts()
                print((self.reset_castle()))
            if e.key == K_z:
                # get_savable_upgrade_amounts()
                print((self.unit.stats))

            # elif e.key == K_s:
            #    r = self.unit.try_use_steam(1)
            #    print "tried to use steam :%s:" % r

    def reset_castle(self):
        """resets the castle stats to the backup."""

        if self.g.level.game.backup_castle_stats:
            print("before")
            print((self.unit.stats))
            self.unit.stats = copy.deepcopy(self.g.level.game.backup_castle_stats)
            print("after")
            print((self.unit.stats))
            if hasattr(self.g, "level"):
                if hasattr(self.g.level, "interface"):
                    self.g.level.interface.new_equipment(self.unit.stats)

    def backup_castle(self):
        """backs the castle stats up."""
        self.g.level.game.backup_castle_stats = copy.deepcopy(self.unit.stats)


def castle_new(g, t, value):
    g.clayer[t.ty][t.tx] = 0
    s = CastleSprite(g.images["castle"], t.rect)
    s.g = g
    # print "castle"
    # print t.rect
    s.groups = g.string2groups("castle")
    s.agroups = g.string2groups("robot")

    s.effect = steam.Effect(32, 1)

    g.castle = s

    g.sprites.append(s)
    if g.level.game.should_restore_stats:
        pass
        # TODO:
        s.reset_castle()
        s.unit.stats["Health"] = 10.0


def castle_hit(g, s, a):
    # screams of smushing go here.
    a.state = "dead"

    g.level.game.sm.Play(random.choice(sound_info.squish))

    # print g
    # print dir(s)
    # print a

    # print "castle hit"

    # test for collision between castle sprite and other types...
    # if the sprite has an item then pop up a dialog asking the
    #  player if they want to pick it up.

    # the player shouldn't be able to move when being asked about coal.

    # NOTE: this is only a todo if we stop the castle from moving
    #  whilst picking things up.
    # TODO: after picking not picking up the item, how long before asking again?
    #  maybe put a timer in there for 3. seconds before asking again.

    # store the item sprite to pick up on click.
    # s.item_sprite = a
    # s.last_hit_coal += 1


def castle_loop(g, s):
    # print "castle loop"

    s.unit.loop()

    # update the steam every second.
    # print g.level.game.ticks_passed

    # tw,th = g.iso_w,g.iso_h

    tw, th = g.iso_w, g.iso_h
    sx, sy = s.rect.centerx // tw, s.rect.centery // th

    s.last_update_steam += g.level.game.elapsed_time

    if s.can_move:
        # s.rect.x += 1

        speed = int(s.unit.stats["Speed"])

        kx, ky = 0, 0

        # put in the joystick movement.
        if kx == 0:
            kx += s.direction_dx
            # s.dx = 0
        if ky == 0:
            ky += s.direction_dy
            # s.dy = 0

        # HACK: we do a little invert here.  because hex movement is weird.
        # we try to make up joy go up on screen.

        if 1:
            if kx == 0 and ky == -1:
                kx = -1
                ky = -1

            elif kx == 0 and ky == 1:
                kx = 1
                ky = 1

            elif kx == 1 and ky == -1:
                kx = 0
                ky = -1

            elif kx == -1 and ky == -1:
                kx = -1
                ky = 0
            elif kx == -1 and ky == 1:
                kx = 0
                ky = 1
            elif kx == 1 and ky == 1:
                kx = 1
                ky = 0

            elif kx == 1 and ky == 0:
                kx = 1
                ky = -1

            elif kx == -1 and ky == 0:
                kx = -1
                ky = 1

        # if ky == 1 and kx == 0:
        #    kx = 1
        #    ky = 1

        # print kx,ky

        keyboard_or_joy_speed_slow = 1
        # if ky == 0 or kx == 0:
        #        if (kx,ky) != (0,0):
        #
        #            if (kx,ky) == (1,1) or (kx,ky) == (-1,-1) or kx == 0 or ky == 0:
        #                keyboard_or_joy_speed_slow = 0

        if (kx, ky) != (0, 0):
            keyboard_or_joy = 1

            tx, ty = sx + kx, sy + ky
            if g.castle_layer[ty][tx] == 0:
                s.path = [(tx, ty)]

        dx, dy = 0, 0
        path = s.path
        while len(path) > 0:
            g.auto_scroll = True
            bx, by = path[0]
            if (bx, by) == (sx, sy):
                path.pop(0)
                continue
            dx, dy = bx - sx, by - sy
            # v = 4
            # s.rect.x += sign(dx)*v
            # s.rect.y += sign(dy)*v
            dx, dy = sign(dx), sign(dy)
            break

        if keyboard_or_joy_speed_slow:
            speed = int(round(speed // 2.0))

        if dx != 0 or dy != 0:
            # HACK: so i can move this guy ! :)
            if 1 or s.unit.try_move():
                s.rect.x += dx * speed
                s.rect.y += dy * speed

            if dx < 0:
                s.setimage(g.images["castle.left"])
            if dx > 0:
                s.setimage(g.images["castle.right"])

        # s.rect.clamp_ip(g.view)
    if s.last_hit_coal:
        # print "last_hit_coal, last"
        if s.last_hit_coal == s.last_loop:
            s.last_loop = 0
            s.last_hit_coal = 0
            s.no_pickup("")
        else:
            s.last_loop = s.last_hit_coal

    # okay, so the robots are scared to death of you...
    a = s
    for r in g.robots:
        rx, ry = r.x, r.y
        ax, ay = a.rect.x, a.rect.y
        dx, dy = rx - ax, ry - ay
        dist = (dx * dx + dy * dy) ** 0.5
        if dist and dist < 32 * 3:
            r.min = min(r._min, r.min + 12.0)
            inc = 16.0
            x, y = rx + dx * inc // dist, ry + dy * inc // dist
            robot.robot_shove(g, r, (x, y))

    s.frame += 1


def cball_new(g, pos, dest, damage=1.0, pressure=16):
    # g.level.game.sm.Play(random.choice(sound_info.cannon))
    # g.level.game.sm.Play(sound_info.cannon.nextone())

    if pressure < 9:
        g.level.game.sm.Play(sound_info.cannon[0], wait=3)
        print("cannon 1 sound")
    elif pressure >= 9 and pressure < 15:
        g.level.game.sm.Play(sound_info.cannon[1], wait=3)
        print("cannon 2 sound")
    elif pressure >= 15 and pressure < 25:
        g.level.game.sm.Play(sound_info.cannon[2], wait=3)
        print("cannon 3 sound")
    elif pressure >= 25:
        print("cannon 4 sound,  two cannon 3 sounds played at once.")
        g.level.game.sm.Play(sound_info.cannon[2], wait=3)
        g.level.game.sm.Play(sound_info.cannon[2], wait=3)

    # Here we use a different graphic for the cannon balls depending on Damage.

    if damage <= 1.0:
        img_name = "cball"
        num_steam, num_steam_add = 20, 5
    elif damage > 1.0 and damage < 4.0:
        img_name = "cball2"
        num_steam, num_steam_add = 30, 10
    elif damage >= 4.0 and damage < 6.0:
        img_name = "cball3"
        num_steam, num_steam_add = 25, 15
    elif damage >= 6.0 and damage < 8.0:
        img_name = "cball4"
        num_steam, num_steam_add = 35, 20
    elif damage >= 8.0 and damage < 10.0:
        img_name = "cball4"
        num_steam, num_steam_add = 42, 23
    elif damage >= 10.0:
        img_name = "cball4"
        num_steam, num_steam_add = 50, 25

    s = isovid.Sprite(g.images[img_name], pos)

    # we assign the damage that is done by this cannon ball.
    s.damage = damage
    s.num_steam, s.num_steam_add = num_steam, num_steam_add

    s.frame = 0

    g.sprites.append(s)
    s.loop = cball_loop
    s.hit = cball_hit
    s.groups = g.string2groups("cball")
    s.agroups = g.string2groups("robot,cannon")  # ,factory,truck,cannon")

    # s.rect, s._rect

    dx, dy = dest[0] - s.rect.x, dest[1] - s.rect.y
    dist = (dx * dx + dy * dy) ** 0.5
    if dist <= 0:
        return
    v = float(pressure)
    s.vx, s.vy = dx * v // dist, dy * v // dist
    s.vz = -5
    s.z = 0

    # we add a steam effect for when firing the cannon ball.
    #   It's steamyness is different depending on the damage of the cannon.
    rect = pygame.Rect(s.rect)
    rect.x += s.vx * 2
    rect.y += s.vy * 2
    effect.effect_new(
        g, rect, steam.Effect(num_steam, num_steam_add, 12, 0, (255, 255, 255)), 20
    )


def cball_loop(g, s):
    try:
        s.rect.x += s.vx
    except AttributeError:
        return

    s.rect.y += s.vy
    s.z += s.vz
    s.vz += 1
    if s.vz == 8:
        if s in g.sprites:
            g.sprites.remove(s)

        img_name = get_hole_for_damage(s.damage)
        g.bkgr_blit(g.images[img_name], (s.rect.centerx, s.rect.centery))

        g.level.game.sm.Play(sound_info.hitground.nextone(), wait=3)

        if HOLES_ARE_TILES:
            tw, th = g.iso_w, g.iso_h
            tx, ty = s.rect.centerx // tw, s.rect.centery // th

            dirs = [(0, 0)]
            if HOLES_ARE_TILES > 1:
                dirs.extend([(-1, 0), (1, 0), (0, 1), (0, -1)])
            for dx, dy in dirs:
                xx, yy = tx + dx, ty + dy
                v = g.get((xx, yy))
                if v in (0, 7):
                    g.set((xx, yy), 30)

        # ss = effect.effect_new(g,s.rect,steam.Effect(20,20,8,0,(117,110,94)),20)
        num_steam, num_steam_add = s.num_steam, s.num_steam_add
        ss = effect.effect_new(
            g,
            s.rect,
            steam.Effect(
                int(num_steam * 2), int(num_steam_add * 2), num_steam_add, 0, (0, 0, 0)
            ),
            20,
        )
        ss.z = 8

        print("gahhh")
        a = s
        for r in g.robots:
            rx, ry = r.x, r.y
            ax, ay = a.rect.x, a.rect.y
            dx, dy = rx - ax, ry - ay
            dist = (dx * dx + dy * dy) ** 0.5
            if dist and dist < 32 * 4:
                r.min = min(r._min, r.min + 12.0)
                inc = 31.0
                x, y = rx + dx * inc // dist, ry + dy * inc // dist
                robot.robot_shove(g, r, (x, y))

            if dist < 32 * 3:  # 1.5:
                r.state = "dead"

    # print s.rect
    s.frame += 1


class SpriteBlast:
    def __init__(self):
        self.e1 = explode.Effect(64, 16)
        # self.e2 = steam.Effect(128,32,24,0)
        self.e2 = steam.Effect(120, 40, 16, 0)

    def loop(self, pos):
        self.e1.loop(pos)
        self.e2.loop(pos)

    def paint(self, screen, origin):
        self.e1.paint(screen, origin)
        self.e2.paint(screen, origin)


def cball_hit(g, a, b):
    # we just destroy 'robots'... aka peasants.  cball keeps going.
    if hasattr(b, "type") and b.type == "robot":
        if b in g.sprites:
            # g.sprites.remove(b)
            b.state = "dead"
    else:
        return cball_hit_old(g, a, b)


def get_hole_for_damage(damage):
    if damage <= 1.0:
        img_name = "hole"
    elif damage > 1.0 and damage < 4.0:
        img_name = "hole2"
    elif damage >= 4.0 and damage < 6.0:
        img_name = "hole3"
    elif damage >= 6.0 and damage < 8.0:
        img_name = "hole4"
    elif damage >= 8.0 and damage < 10.0:
        img_name = "hole4"
    elif damage >= 10.0:
        img_name = "hole4"
    return img_name


def cball_hit_old(g, a_cannon_ball, b):
    ss = effect.effect_new(g, b.rect, steam.Effect(60, 20, 4, 0), 20)

    # lets hit the unit, and see if it dies.
    if b.unit.hit(a_cannon_ball.damage):
        g.level.game.sm.Play(sound_info.destroyenemy.nextone())

        #         if b.unit.name == "factory":
        #             g.level.game.state = states.NextLevel(g.level.game, g.level.game.state)
        #             print "you won!"
        #             return

        if b in g.sprites:
            g.sprites.remove(b)

        img_name = get_hole_for_damage(a_cannon_ball.damage)

        g.bkgr_blit(
            g.images[img_name], (a_cannon_ball.rect.centerx, a_cannon_ball.rect.centery)
        )

        # do an explosion here...
        # that looks like...
        ss = effect.effect_new(g, b.rect, SpriteBlast(), 20)
    else:
        # we only play the hit sound if the thing doesn't die?
        g.level.game.sm.Play(sound_info.hitenemy.nextone())

    try:
        g.sprites.remove(a_cannon_ball)
    except ValueError:
        # already removed??? yeah that happens :)
        # if the ball hits more than one tile
        pass

    print(("hit!", b.unit.name))
    pass



# i will take over the world
# zanthor shall rule the universe
# my powers can not be stopped
# my strength has no bounds
# die worthless scum
# My name is Zanthor.  I like flowers, and killing.
# fear me weak humans
# When Zanthor was little(that's me) I always wanted to grow up and be a junkie or a telemarketer.  Unfortunately it didn't turn out that way, and instead I have become a dark overlord.


data = {
    "SENTENCE": [
        "[_SENTENCE]",
    ],
    "_SENTENCE": [
        "[NAME] [WILL] [KILL] [HUMANS].",
        "[FEAR] [NAME].",
        "[HUMANS], [FEAR] [NAME].",
        "[FEAR] [NAME], [HUMANS].",
        "[NAME] [WILL] [KILL] and [KILL] [HUMANS].",
        "[ITEMS] [KILL] [ITEMS] [KILL].",
        "[NAME] [LIKES] [ITEMS].",
        "My name is [NAME].  I [LIKE] [ITEMS].",
        "My name is [NAME].  I [LIKE] [ITEMS] and [KILLING].",
        "I have become [NAME].",
        "I once was [WEAK].  I am now [NAME].",
        "I am [NAME].",
        "I am [NAME].  Hear me [KILL].",
        "I [KILL] [HUMANS].",
        "[HUMANS] will become my [SLAVES].",
    ],
    "NAME": [
        "[_NAME]",
        "[_NAME] the [GREAT]",
        "[_NAME] the [VERY] [GREAT]",
        "[_NAME], [KING] of [HUMANS]",
    ],
    "_NAME": [
        "Zanthor",
        "Lord Zanthor",
        "Evil Lord Zanthor",
    ],
    "VERY": [
        "very",
        "exceedingly",
        "abundantly",
        "amazingly",
        "indisputably",
        "amazingly",
        "awesomely",
        "shatteringly",
        "freighteningly",
        "consumingly",
    ],
    "GREAT": [
        "awesome",
        "great",
        "all-mightly",
        "powerful",
        "magnificent",
        "regal",
        "rotund",
        "wise",
        "magestic",
        "stately",
        "exalted",
    ],
    "KING": [
        "king",
        "ruler",
        "master",
        "owner",
        "magistrate",
        "conqueror",
        "captain",
        "dictator",
        "president",
        "potentate",
        "overlord",
    ],
    "HUMANS": [
        "[_HUMANS]",
        "[WEAK] [_HUMANS]",
        "[WEAK] and [WEAK] [_HUMANS]",
    ],
    "WEAK": [
        "weak",
        "pathetic",
        "feable",
        "ill-bred",
        "sorry",
        "unworthy",
        "poor",
        "unfortunate",
        "sad",
        "lost",
        "ugly",
        "disgusting",
        "hideous",
        "sorry",
        "unwelcome",
        "measly",
    ],
    "_HUMANS": [
        "humans",
        "people",
        "sheep",
        "masses",
        "scum",
        "earthlings",
        "citizens",
        "fools",
    ],
    "WILL": [
        "will",
        "plans to",
        "aims to",
        "must",
        "shall",
        "can easily",
        "will always",
    ],
    "LIKES": [
        "likes",
        "loves",
        "covets",
        "collects",
        "eats",
        "hordes",
    ],
    "LIKE": [
        "like",
        "love",
        "covet",
        "collect",
        "eat",
        "horde",
    ],
    "FEAR": [
        "fear",
        "loath",
        "tremble before",
        "shy away from",
        "hide from",
        "run from",
        "faint near",
    ],
    "KILL": [
        "kill",
        "maim",
        "batter",
        "harm",
        "hurt",
        "maltreat",
        "mangle",
        "wound",
        "maul",
        "ruin",
        "smite",
        "break",
        "punish",
        "blow-up",
        "blast",
        "ransack",
        "destroy",
    ],
    "KILLING": [
        "killing",
        "maiming",
        "battering",
        "harming",
        "hurting",
        "maltreating",
        "mangling",
        "wounding",
        "mauling",
        "ruining",
        "smiting",
        "breaking",
        "punishing",
        "blowing-up",
        "blasting",
        "ransacking",
        "destroying",
    ],
    "SLAVES": [
        "slaves",
        "underlings",
        "footstool",
        "minions",
        "onions",
        "breakfast",
        "flower pot mulch",
        "belly button lint",
        "toe jam",
    ],
    "ITEMS": [
        "[_ITEMS]",
        "[_ITEMS]",
        "[_ITEMS] and [_ITEMS]",
    ],
    "_ITEMS": [
        "coal",
        "water",
        "cannons",
        "power",
        "steam",
        "destruction",
        "blasts",
        "flowers",
        "explosions",
        "fire",
        "desolation",
    ],
}

import random
import re


def replace(v):
    return random.choice(data[v.group(1)])


def ureplace(v):
    return v.group(0).upper()


def generate():
    text = random.choice(data["SENTENCE"])
    regex = re.compile("\[(\w+)\]")
    while "[" in text:
        text = re.sub(regex, replace, text, 1)
    regex = re.compile("[\.\!\?]\s+\w")
    text = re.sub(regex, ureplace, text)
    return text[0:1].upper() + text[1:]


def generate_upgrade_message(upgrade_what):
    return "Zanthor upgrades the %s on his castle." % (upgrade_what.replace("Up", ""))


def generate_win_level():
    return generate()


if __name__ == "__main__":
    for n in range(0, 12):
        print((generate()))
        print("")



"""A timer for games with set-rate FPS.
"""

import pygame


class Timer:
    """A timer for games with set-rate FPS.

    <pre>Timer(fps)</pre>
    """

    def __init__(self, fps):
        self.wait = 1000 / fps
        self.nt = pygame.time.get_ticks()

    def tick(self):
        """Wait correct amount of time each frame.  Call this once per frame.

        <pre>Timer.tick()</pre>
        """
        self.ct = pygame.time.get_ticks()
        if self.ct < self.nt:
            pygame.time.wait(self.nt - self.ct)
            self.nt += self.wait
        else:
            self.nt = pygame.time.get_ticks() + self.wait


class Speedometer:
    """A timer replacement that returns out FPS once a second.
    <pre>Speedometer()</pre>

    <strong>Attributes</strong>
    <dl>
    <dt>fps <dd>always set to the current FPS
    </dl>
    """

    def __init__(self):
        self.frames = 0
        self.st = pygame.time.get_ticks()
        self.fps = 0

    def tick(self):
        """Call this once per frame.

        <pre>Speedometer.tick()</pre>
        """
        r = None
        self.frames += 1
        self.ct = pygame.time.get_ticks()
        if (self.ct - self.st) >= 1000:
            r = self.fps = self.frames
            # print "%s: %d fps"%(self.__class__.__name__,self.fps)
            self.frames = 0
            self.st += 1000
        pygame.time.wait(0)  # NOTE: not sure why, but you gotta call this now and again
        return r


# vim: set filetype=python sts=4 sw=4 noet si :



import pygame

from .const import *
from . import table
from . import group
from . import button, basic


def action_open(value):
    print("gui.action_open", "Scheduled to be deprecated.")
    value.setdefault("x", None)
    value.setdefault("y", None)
    value["container"].open(value["window"], value["x"], value["y"])


def action_setvalue(value):
    print("gui.action_setvalue", "Scheduled to be deprecated.")
    a, b = value
    b.value = a.value


def action_quit(value):
    print("gui.action_quit", "Scheduled to be deprecated.")
    value.quit()


def action_exec(value):
    print("gui.action_exec", "Scheduled to be deprecated.")
    exec(value["script"], globals(), value["dict"])


class Toolbox(table.Table):
    def __setattr__(self, k, v):
        _v = self.__dict__.get(k, NOATTR)
        self.__dict__[k] = v
        if k == "value" and _v != NOATTR and _v != v:
            self.group.value = v
            for w in self.group.widgets:
                if w.value != v:
                    w.pcls = ""
                else:
                    w.pcls = "down"
            self.repaint()

    def _change(self, value):
        self.value = self.group.value
        self.send(CHANGE)

    def __init__(self, data, cols=0, rows=0, tool_cls="tool", value=None, **params):
        print("gui.Toolbox", "Scheduled to be deprecated.")
        params.setdefault("cls", "toolbox")
        table.Table.__init__(self, **params)

        if cols == 0 and rows == 0:
            cols = len(data)
        if cols != 0 and rows != 0:
            rows = 0

        self.tools = {}

        _value = value

        g = group.Group()
        self.group = g
        g.connect(CHANGE, self._change, None)
        self.group.value = _value

        x, y, p, s = 0, 0, None, 1
        for ico, value in data:
            # from __init__ import theme
            from . import app

            img = app.App.app.theme.get(tool_cls + "." + ico, "", "image")
            if img:
                i = basic.Image(img)
            else:
                i = basic.Label(ico, cls=tool_cls + ".label")
            p = button.Tool(g, i, value, cls=tool_cls)
            self.tools[ico] = p
            # p.style.hexpand = 1
            # p.style.vexpand = 1
            self.add(p, x, y)
            s = 0
            if cols != 0:
                x += 1
            if cols != 0 and x == cols:
                x, y = 0, y + 1
            if rows != 0:
                y += 1
            if rows != 0 and y == rows:
                x, y = x + 1, 0



"""a collection of text rendering functions
"""


def write(s, font, pos, color, text, border=1):
    """write text to a surface with a black border

    <pre>write(s,font,pos,color,text,border=1)</pre>
    """
    i = font.render(text, 1, (0, 0, 0))
    si = border
    dirs = [(-1, -1), (-1, 0), (-1, 1), (0, -1), (0, 1), (1, -1), (1, 0), (1, 1)]
    for dx, dy in dirs:
        s.blit(i, (pos[0] + dx * si, pos[1] + dy * si))
    i = font.render(text, 1, color)
    s.blit(i, pos)


def writec(s, font, color, text, border=1):
    """write centered text to a surface with a black border

    <pre>writec(s,font,color,text,border=1)</pre>
    """
    w, h = font.size(text)
    x = (s.get_width() - w) / 2
    y = (s.get_height() - h) / 2
    write(s, font, (x, y), color, text, border)


def writepre(s, font, rect, color, text):
    """write preformatted text

    <pre>writepre(s,font,rect,color,text)</pre>
    """
    r, c, txt = rect, color, text
    txt = txt.replace("\t", "        ")
    i = font.render(" ", 1, c)
    sw, sh = i.get_width(), i.get_height()
    y = r.top
    for sentence in txt.split("\n"):
        x = r.left
        i = font.render(sentence, 1, c)
        s.blit(i, (x, y))
        y += sh


def writewrap(s, font, rect, color, text):
    """write wrapped text

    <pre>writewrap(s,font,rect,color,text)</pre>
    """
    r, c, txt = rect, color, text
    txt = txt.replace("\t", "        ")
    i = font.render(" ", 1, c)
    sw, sh = i.get_width(), i.get_height()
    y = r.top
    for sentence in txt.split("\n"):
        x = r.left
        for word in sentence.split(" "):
            i = font.render(word, 1, c)
            iw, ih = i.get_width(), i.get_height()
            if x + iw > r.right:
                x, y = r.left, y + sh
            s.blit(i, (x, y))
            x += iw + sw
        y += sh


# vim: set filetype=python sts=4 sw=4 noet si :



from . import isovid
import random

from .const import *

from .units import CannonTower

from . import sound_info


def cannon_new(g, t, v):
    g.clayer[t.ty][t.tx] = 0
    s = isovid.Sprite(g.images["cannon"], t.rect)

    s.frame = random.randrange(0, FPS * 3)

    g.sprites.append(s)
    s.loop = cannon_loop
    s.hit = cannon_hit
    s.groups = g.string2groups("cannon")
    # s.agroups =g.string2groups("castle")

    s.unit = CannonTower()


def cannon_loop(g, s):
    s.unit.loop()

    tw, th = g.iso_w, g.iso_h
    if s.frame % (FPS * 3) == 0:
        sx, sy = s.rect.centerx / tw, s.rect.centery / th
        p = g.castle
        px, py = p.rect.centerx / tw, p.rect.centery / th
        dx, dy = px - sx, py - sy
        if (dx * dx + dy * dy) ** 0.5 < 12:  # only shoot if they are reachable
            # try and use some steam to fire.
            if s.unit.try_fire():
                cball_new(g, (s.rect.centerx, s.rect.centery), s.unit.stats["Damage"])

    s.frame += 1


def cannon_hit(g, a, b):
    pass


def cball_new(g, pos, damage=1.0):
    g.level.game.sm.Play("cannon")

    s = isovid.Sprite(g.images["cball"], pos)

    s.frame = 0

    g.sprites.append(s)
    s.loop = cball_loop
    s.hit = cball_hit
    # s.groups = g.string2groups("cball")
    s.agroups = g.string2groups("castle")

    # s.rect, s._rect

    dx, dy = g.castle.rect.x - s.rect.x, g.castle.rect.y - s.rect.y
    dist = (dx * dx + dy * dy) ** 0.5
    v = 16
    s.vx, s.vy = dx * v / dist, dy * v / dist
    s.vz = -8
    s.z = 0
    s.damage = damage


def cball_loop(g, s):
    s.rect.x += s.vx
    s.rect.y += s.vy
    s.z += s.vz
    s.vz += 1
    if s.vz == 8:
        g.sprites.remove(s)
    # print s.rect
    s.frame += 1


def cball_hit(g, a, b):
    print("BANG!")
    g.level.game.sm.Play(sound_info.ushit.nextone())

    if b.unit.hit(a.damage):
        # raise "game over, you got killed!"
        # b.reset_castle()
        from . import states
        g.level.game.state = states.GameOver(g.level.game)  # , g.level.game.state)

    pass
